
/*
 * Copyright (c) Abraham vd Merwe <abz@blio.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	  notice, this list of conditions and the following disclaimer in the
 *	  documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of other contributors
 *	  may be used to endorse or promote products derived from this software
 *	  without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pwd.h>
#include <grp.h>
#include <errno.h>
#include <string.h>

#if defined(GETHOSTBYNAME) || defined(GETSERVBYNAME)
#include <netdb.h>
#endif	/* #if defined(GETHOSTBYNAME) || defined(GETSERVBYNAME) */

#include <debug/memory.h>
#include <debug/log.h>

#include <config/config.h>

#include <tinysnmp/compat/inttypes.h>

#include <abz/aton.h>
#include <abz/atou16.h>

#include "config.h"

static config_t config;

/*
 * free all memory allocated by config_parse().
 */
static void config_destroy ()
{
   mem_free (config.community);
   mem_free (config.modules);
   mem_free (config.confdir);
   mem_free (config.allow);
   mem_free (config.pidfile);
}

/*
 * parse configuration file. return a pointer to a static configuration
 * structure if successful, NULL if some error occurred. the function
 * automatically logs all errors.
 */
const config_t *config_parse (const char *filename)
{
   char *s,**sa;
   bool b;
   int i,n;
   long int li;
#ifdef GETSERVBYNAME
   struct servent *serv;
#endif	/* #ifdef GETSERVBYNAME */
   struct passwd *pw;
   struct group *gr;

   if (!cfg_open (filename)) return (NULL);

   if (!cfg_get_str (&config.community,"community"))
	 {
		log_printf (LOG_ERROR,"%s: No ``community'' string defined!\n",filename);
		goto error0;
	 }

   if (!cfg_get_str (&config.modules,"modules"))
	 {
		log_printf (LOG_ERROR,"%s: No ``modules'' string defined!\n",filename);
		goto error1;
	 }

   if (!cfg_get_str (&config.confdir,"confdir"))
	 {
		log_printf (LOG_ERROR,"%s: No ``confdir'' string defined!\n",filename);
		goto error2;
	 }

   config.addr.sin_family = PF_INET;
   config.addr.sin_addr.s_addr = INADDR_ANY;
   config.addr.sin_port = htons (161);

   if (cfg_get_str (&s,"listen"))
	 {
		struct in_addr addr;
#ifdef GETHOSTBYNAME
		struct hostent *host;
#endif	/* #ifdef GETHOSTBYNAME */

		if (inet_aton (s,&addr))
		  {
			 config.addr.sin_addr.s_addr = addr.s_addr;
			 mem_free (s);
		  }
#ifdef GETHOSTBYNAME
		else if ((host = gethostbyname (s)) != NULL)
		  {
			 memcpy (&config.addr.sin_addr,(struct in_addr *) host->h_addr,sizeof (struct in_addr));
			 mem_free (s);
		  }
#endif	/* #ifdef GETHOSTBYNAME */
		else
		  {
#ifdef GETHOSTBYNAME
			 log_printf (LOG_ERROR,"%s: Invalid hostname/ip-address %s!\n",filename,s);
#else	/* #ifdef GETHOSTBYNAME */
			 log_printf (LOG_ERROR,"%s: Invalid ip-address %s!\n",filename,s);
#endif	/* #ifdef GETHOSTBYNAME */
			 mem_free (s);
			 goto error3;
		  }
	 }

   if (cfg_get_str (&s,"port"))
	 {
		uint16_t port;

		if (!atou16 (s,&port))
		  {
			 mem_free (s);
			 config.addr.sin_port = htons (port);
		  }
#ifdef GETSERVBYNAME
		else if ((serv = getservbyname (s,"udp")) != NULL)
		  {
			 mem_free (s);
			 config.addr.sin_port = serv->s_port;
		  }
#endif	/* #ifdef GETSERVBYNAME */
		else
		  {
#ifdef GETSERVBYNAME
			 log_printf (LOG_ERROR,"%s: Invalid service/port number %s!\n",filename,s);
#else	/* #ifdef GETSERVBYNAME */
			 log_printf (LOG_ERROR,"%s: Invalid port number %s!\n",filename,s);
#endif	/* #ifdef GETSERVBYNAME */
			 mem_free (s);
			 goto error3;
		  }
	 }
#ifdef GETSERVBYNAME
   else if ((serv = getservbyname ("snmp","udp")) != NULL)
	 config.addr.sin_port = serv->s_port;
#endif	/* #ifdef GETSERVBYNAME */

   if (!cfg_get_str_array (&sa,&n,"allow"))
	 {
		log_printf (LOG_ERROR,"%s: No ``allow'' string list defined!\n",filename);
		goto error3;
	 }

   if ((config.allow = (network_t *) mem_alloc (sizeof (network_t) * n)) == NULL)
	 {
		log_printf (LOG_ERROR,"Out of memory!\n");
		for (i = 0; i < n; i++) mem_free (sa[i]);
		mem_free (sa);
		goto error3;
	 }

   for (i = 0; i < n; i++)
	 if (aton (&config.allow[i],sa[i]) < 0)
	   {
		  log_printf (LOG_ERROR,"%s: Invalid network %s defined in list of allowed networks!\n",filename,sa[i]);
		  for (i = 0; i < n; i++) mem_free (sa[i]);
		  mem_free (sa);
		  goto error4;
	   }

   for (i = 0; i < n; i++) mem_free (sa[i]);
   mem_free (sa);

   config.n = n;

   if (!cfg_get_bool (&b,"reuseaddr"))
	 {
		log_printf (LOG_ERROR,"%s: ``reuseaddr'' boolean value not found!\n",filename);
		goto error4;
	 }

   config.reuseaddr = b;

   if (cfg_get_str (&s,"user"))
	 {
		if ((pw = getpwnam (s)) == NULL)
		  {
			 log_printf (LOG_ERROR,"%s: Couldn't convert user name %s to a uid: %s\n",filename,s,strerror (errno));
			 mem_free (s);
			 goto error4;
		  }
		config.uid = pw->pw_uid;
		mem_free (s);
	 }
   else if (cfg_get_int (&li,"user"))
	 {
		if (li < 0)
		  {
			 log_printf (LOG_ERROR,"%s: ``user'' value out of range!\n",filename);
			 goto error4;
		  }
		config.uid = li;
	 }
   else
	 {
		log_printf (LOG_ERROR,"%s: No ``user'' string/integer found!\n",filename);
		goto error4;
	 }

   if (cfg_get_str (&s,"group"))
	 {
		if ((gr = getgrnam (s)) == NULL)
		  {
			 log_printf (LOG_ERROR,"%s: Couldn't convert group name %s to a gid: %s\n",filename,s,strerror (errno));
			 mem_free (s);
			 goto error4;
		  }
		config.gid = gr->gr_gid;
		mem_free (s);
	 }
   else if (cfg_get_int (&li,"group"))
	 {
		if (li < 0)
		  {
			 log_printf (LOG_ERROR,"%s: ``group'' value out of range!\n",filename);
			 goto error4;
		  }
		config.gid = li;
	 }
   else
	 {
		log_printf (LOG_ERROR,"%s: No ``group'' string/integer found!\n",filename);
		goto error4;
	 }

   if (!cfg_get_str (&config.pidfile,"pidfile"))
	 {
		log_printf (LOG_ERROR,"%s: No ``pidfile'' string found!\n",filename);
		goto error4;
	 }

   cfg_close ();
   atexit (config_destroy);
   return (&config);

error4:
   mem_free (config.allow);

error3:
   mem_free (config.confdir);

error2:
   mem_free (config.modules);

error1:
   mem_free (config.community);

error0:
   cfg_close ();
   return (NULL);
}

