
/*
 * Copyright (c) Abraham vd Merwe <abz@blio.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	  notice, this list of conditions and the following disclaimer in the
 *	  documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of other contributors
 *	  may be used to endorse or promote products derived from this software
 *	  without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <dirent.h>
#include <dlfcn.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>
#include <limits.h>

#include <ber/ber_utils.h>
#include <debug/memory.h>
#include <debug/log.h>

#include <tinysnmp/agent/odb.h>
#include <tinysnmp/agent/mod.h>

#include "mod.h"

typedef struct child_r
{
   const mod_t *mod;		/* module data */
   int dirty;				/* have update() function been called yet? */
   struct child_r *next;
} child_t;

typedef struct parent_r
{
   void *dl;				/* module handle used by dlsym(), dlclose(), etc. */
   struct parent_r *next;
} parent_t;

typedef struct
{
   parent_t *parents;		/* list of modules (files) */
   child_t *children;		/* list of modules (exported by parent modules) */
} mod_private_t;

static mod_private_t mod;

static void print_dl_error (const char *function,const char *filename)
{
   const char *error = dlerror ();

   log_printf (LOG_WARNING,"while loading library %s:",filename);
   if (error != NULL)
	 log_printf (LOG_WARNING,"%s: %s\n",function,error);
   else
	 log_printf (LOG_WARNING,"%s failed\n",function);
}

static int oidcmp (const uint32_t *a,const uint32_t *b)
{
   uint32_t min = a[0] < b[0] ? a[0] : b[0];
   int i;

   for (i = 1; i <= min; i++)
	 {
		if (a[i] < b[i]) return (-1);
		if (a[i] > b[i]) return (1);
	 }

   return (0);
}

static void *load_library (child_t **children,const char *filename,const char *confdir)
{
   int i,n = 0;
   void *dl;
   const mod_t *mod;
   child_t *child,*cur,*prev;

   if ((dl = dlopen (filename,RTLD_NOW)) == NULL)
	 {
		print_dl_error ("dlopen()",filename);
		return (NULL);
	 }

   if ((mod = dlsym (dl,"mod")) == NULL)
	 {
		print_dl_error ("dlsym()",filename);
		dlclose (dl);
		return (NULL);
	 }

   for (i = 0; mod[i].name != NULL; i++)
	 {
		if (mod[i].update == NULL)
		  {
			 log_printf (LOG_WARNING,"while loading library %s:\n"
						 "    module %s: update() function is null\n"
						 "    not loading module %s\n",
						 filename,
						 mod[i].name,
						 mod[i].name);
			 continue;
		  }

		if (mod[i].oid == NULL)
		  {
			 log_printf (LOG_WARNING,
						 "while loading library %s:\n"
						 "    module %s: exported oid range is null\n"
						 "    not loading module %s\n",
						 filename,
						 mod[i].name,
						 mod[i].name);
			 continue;
		  }

		if ((child = (child_t *) mem_alloc (sizeof (child_t))) == NULL)
		  {
			 log_printf (LOG_WARNING,
						 "while loading library %s:\n"
						 "    module %s: %s\n"
						 "    not loading module %s\n",
						 filename,
						 mod[i].name,
						 strerror (errno),
						 mod[i].name);
			 continue;
		  }

		if (mod[i].open != NULL && mod[i].open (&mod[i],confdir) < 0)
		  {
			 mem_free (child);
			 continue;
		  }

		child->mod = &mod[i];
		child->dirty = 0;
		child->next = NULL;

		if (*children != NULL)
		  {
			 for (cur = *children, prev = NULL; cur != NULL; prev = cur, cur = cur->next)
			   if (oidcmp (cur->mod->oid,child->mod->oid) > 0)
				 break;

			 if (cur != NULL)
			   {
				  child->next = cur;
				  if (prev != NULL) prev->next = child; else *children = child;
			   }
			 else prev->next = child;
		  }
		else *children = child;

		log_printf (LOG_NORMAL,"registered module %s\n",child->mod->name);

		n++;
	 }

   if (!n)
	 {
		log_printf (LOG_WARNING,"while loading library %s: no modules found\n",filename);
		dlclose (dl);
		return (NULL);
	 }

   return (dl);
}

static void child_destroy (child_t **child)
{
   child_t *tmp;

   while (*child != NULL)
	 {
		tmp = *child;
		*child = (*child)->next;
		if (tmp->mod->close != NULL) tmp->mod->close (tmp->mod);
		mem_free (tmp);
	 }
}

static void parent_destroy (parent_t **parent)
{
   parent_t *tmp;

   while (*parent != NULL)
	 {
		tmp = *parent;
		*parent = (*parent)->next;
//		dlclose (tmp->dl);
		mem_free (tmp);
	 }
}

static void mod_close ()
{
   child_destroy (&mod.children);
   parent_destroy (&mod.parents);
   odb_destroy ();
}

static int parent_save (parent_t **parent,void *dl)
{
   parent_t *tmp;

   if ((tmp = (parent_t *) mem_alloc (sizeof (parent_t))) == NULL)
	 return (-1);

   tmp->dl = dl;
   tmp->next = *parent;
   *parent = tmp;

   return (0);
}

/*
 * Open all modules found in module path path''. All regular files ending
 * with .so extentions are considered modules.
 *
 * Returns 0 if successful, or -1 if no modules could be loaded. The function
 * logs all errors and warnings automatically via the log subsystem.
 */
extern int mod_open (const char *path,const char *confdir)
{
   DIR *dir;
   struct dirent *entry;
   char filename[PATH_MAX + 1];
   size_t n = 0,pathlen = strlen (path) + 1;
   void *dl;
   struct stat buf;

   if (pathlen >= PATH_MAX)
	 {
		log_printf (LOG_ERROR,"%s too long\n",path);
		return (-1);
	 }

   if ((dir = opendir (path)) == NULL)
	 {
		log_printf (LOG_ERROR,"opendir(%s): %s\n",path,strerror (errno));
		odb_destroy ();
		return (-1);
	 }


   mod.parents = NULL;
   mod.children = NULL;

   strcpy (filename,path);
   strcat (filename,"/");

   while ((entry = readdir (dir)) != NULL)
	 {
		if (strcmp (entry->d_name + strlen (entry->d_name) - 3,".so"))
		  continue;

		if (pathlen + strlen (entry->d_name) > PATH_MAX)
		  {
			 log_printf (LOG_WARNING,"not loading library %s/%s: filename too long\n",path,entry->d_name);
			 continue;
		  }

		strcpy (filename + pathlen,entry->d_name);

		if (stat (filename,&buf) < 0)
		  {
			 log_printf (LOG_WARNING,"not loading library %s: stat(): %s\n",filename,strerror (errno));
			 continue;
		  }

		dl = NULL;

		if (S_ISREG (buf.st_mode) && ((dl = load_library (&mod.children,filename,confdir)) == NULL || parent_save (&mod.parents,dl) < 0))
		  {
			 log_printf (LOG_WARNING,"not loading library %s: %s\n",filename,strerror (errno));
			 if (dl != NULL)
			   {
				  dlclose (dl);
				  dl = NULL;
			   }
		  }

		if (dl != NULL)
		  {
			 log_printf (LOG_NORMAL,"successfully loaded library %s\n",filename);
			 n++;
		  }
	 }

   closedir (dir);

   if (!n)
	 {
		log_printf (LOG_WARNING,"no libraries found\n");
		mod_close ();
		return (-1);
	 }

   atexit (mod_close);

   return (0);
}

/*
 * Clear the dirty flags on all loaded modules. If these flags are set,
 * no ObjectID's will be updated by the mod_update() function.
 */
void mod_clear ()
{
   child_t *child = mod.children;

   while (child != NULL)
	 {
		if (child->dirty != time (NULL)) child->dirty = 0;
		child = child->next;
	 }
}

static int oidsub (const uint32_t *haystack,const uint32_t *needle)
{
   uint32_t min = haystack[0] < needle[0] ? haystack[0] : needle[0];
   int i;

   for (i = 1; i <= min; i++)
	 if (haystack[i] < needle[i] || haystack[i] > needle[i])
	   return (0);

   return (haystack[0] <= needle[i]);
}

/*
 * If the specified ObjectID falls into one of the ObjectID ranges
 * which are exported by any of the loaded modules, and if that module's
 * dirty flag is not set, update that exported ObjectID range.
 *
 * Returns 0 if successful, -1 if some error occurred. The function
 * logs all errors and warnings automatically via the log subsystem.
 */
int mod_update (const uint32_t *oid)
{
   child_t *child = mod.children;

   while (child != NULL)
	 {
		if (!child->dirty && oidsub (oid,child->mod->oid))
		  {
			 child->dirty = time (NULL);
			 return (child->mod->update (child->mod));
		  }
		child = child->next;
	 }

   return (0);
}

/*
 * If the specified ObjectID falls into one of the ObjectID ranges
 * which are exported by any of the loaded modules, then (if the
 * dirty flags of the modules in question are not set), then the
 * ObjectID's of that module and the one that lexographically succeeds
 * it (if any) are also updated.
 *
 * Returns 0 if successful, -1 if some error occurred. The function
 * logs all errors and warnings automatically via the log subsystem.
 */
int mod_update_next (const uint32_t *oid)
{
   child_t *cur,*prev;

   for (cur = mod.children, prev = NULL; cur != NULL; prev = cur, cur = cur->next)
	 if (oidcmp (cur->mod->oid,oid) >= 0)
	   break;

   if (cur != NULL && cur->next != NULL)
	 {
		if (!cur->next->dirty)
		  {
			 if (cur->next->mod->update (cur->next->mod) < 0)
			   return (-1);

			 cur->next->dirty = time (NULL);
		  }
	 }

   if (cur != NULL)
	 {
		if (!cur->dirty)
		  {
			 if (cur->mod->update (cur->mod) < 0)
			   return (-1);

			 cur->dirty = time (NULL);
		  }
	 }

   if (prev != NULL)
	 {
		if (!prev->dirty)
		  {
			 if (prev->mod->update (prev->mod) < 0)
			   return (-1);

			 prev->dirty = time (NULL);
		  }
	 }

   return (0);
}

/*
 * Define this to 1 if you want to test the module interface code.
 */

#if 0
static void fail (const char *fmt, ...)
{
   va_list ap;
   va_start (ap,fmt);
   if (fmt != NULL) log_vprintf (LOG_ERROR,fmt,ap);
   va_end (ap);
   exit (EXIT_FAILURE);
}

static void closeall ()
{
   mem_close ();
   log_close ();
}

static void test_oid (const char *a,const char *b)
{
   uint32_t *oid1 = NULL,*oid2;
   static const char *result[] = { "less than", NULL, "greater than" };
   int tmp;

   if ((oid1 = make_oid (a)) == NULL || (oid2 = make_oid (b)) == NULL)
	 {
		log_printf (LOG_ERROR,"make_oid: %s\n",strerror (errno));
		if (oid1 != NULL) mem_free (oid1);
		exit (EXIT_FAILURE);
	 }

   if ((tmp = oidcmp (oid1,oid2)))
	 log_printf (LOG_DEBUG,"%s is %s %s\n",a,result[tmp + 1],b);
   else
	 log_printf (LOG_ERROR,"either %s is equal to %s or one is a subset of the other\n",a,b);

   mem_free (oid1);
   mem_free (oid2);
}

int main ()
{
   mem_open (fail);
   log_open (NULL,LOG_NOISY,LOG_HAVE_COLORS | LOG_PRINT_FUNCTION);
   atexit (closeall);

   log_printf (LOG_NORMAL,"equal cases\n");
   test_oid ("1.2.3","1.2.3.4");
   test_oid ("1.2.3","1.2.3");
   test_oid ("1.2.3.4","1.2.3");

   log_printf (LOG_NORMAL,"less than cases\n");
   test_oid ("1.2.3","2.1.0");
   test_oid ("1.1.1","1.1.2");
   test_oid ("1.2.2","1.2.3.1");

   log_printf (LOG_NORMAL,"greater than cases\n");
   test_oid ("2.1","1.1");
   test_oid ("1.2.3","1.1");
   test_oid ("1.2.3.4","1.1.2.1");

   exit (EXIT_SUCCESS);
}
#endif

