
/*
 * Copyright (c) Abraham vd Merwe <abz@blio.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	  notice, this list of conditions and the following disclaimer in the
 *	  documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of other contributors
 *	  may be used to endorse or promote products derived from this software
 *	  without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>

#include <debug/memory.h>
#include <debug/log.h>

#include <ber/ber.h>

#include <tinysnmp/tinysnmp.h>
#include <tinysnmp/agent/odb.h>
#include <tinysnmp/compat/inttypes.h>

#include "tree.h"

static tree_t *odb = NULL;

static void purge (void *data)
{
   snmp_value_t *value = (snmp_value_t *) data;

   if (value->type == BER_OCTET_STRING && value->data.OCTET_STRING.len)
	 mem_free (value->data.OCTET_STRING.buf);
   else if (value->type == BER_OID)
	 mem_free (value->data.OID);

   mem_free (value);
}

/*
 * Destroy an ObjectID database. When this function returns, the
 * database is in exactly the same state as when initialized.
 */
void odb_destroy ()
{
   tree_destroy (&odb,purge);
}

/*
 * Add a new ObjectID to the ObjectID database. Returns 0 if
 * successful, -1 otherwise. Check errno to see what error
 * occurred. The following errors is possible:
 *
 *       ENOMEM     Out of memory
 *       EEXIST     An ObjectID like this (or superset of) already exist
 */
int odb_add_stub (const char *filename,int line,const char *function,const uint32_t *oid,const snmp_value_t *value)
{
   branch_t branch;
   snmp_value_t *tmp;

   if ((tmp = (snmp_value_t *) mem_alloc (sizeof (snmp_value_t))) == NULL)
	 {
		err0:
		errno = ENOMEM;
		return (-1);
	 }

   memcpy (tmp,value,sizeof (snmp_value_t));

   if (value->type == BER_OCTET_STRING)
	 {
		tmp->data.OCTET_STRING.len = value->data.OCTET_STRING.len;

		if (value->data.OCTET_STRING.len)
		  {
			 if ((tmp->data.OCTET_STRING.buf = (uint8_t *) mem_alloc (value->data.OCTET_STRING.len)) == NULL)
			   {
				  err1:
				  mem_free (tmp);
				  goto err0;
			   }

			 memcpy (tmp->data.OCTET_STRING.buf,value->data.OCTET_STRING.buf,value->data.OCTET_STRING.len);
		  }
	 }
   else if (value->type == BER_OID)
	 {
		if ((tmp->data.OID = (uint32_t *) mem_alloc ((value->data.OID[0] + 1) * sizeof (uint32_t))) == NULL)
		  goto err1;

		memcpy (tmp->data.OID,value->data.OID,(value->data.OID[0] + 1) * sizeof (uint32_t));
	 }

   branch.branch = (uint32_t *) oid;
   branch.value = (void *) tmp;

   if (tree_add_stub (filename,line,function,&odb,&branch) < 0)
	 {
		purge (branch.value);
		return (-1);
	 }

   return (0);
}

/*
 * Update the value of an existing branch. Returns 0 if successful,
 * or -1 if the ObjectID doesn't exist.
 */
int odb_update (const uint32_t *oid,snmp_value_t *value)
{
   branch_t branch = { branch: (uint32_t *) oid, value: (void *) value };
   void *old;

   if ((old = tree_update (odb,&branch)) == NULL)
	 return (-1);

   purge (old);

   return (0);
}

/*
 * Find an ObjectID in the ObjectID database. Return the value of
 * an existing ObjectID in the database if successful, or NULL if
 * the ObjectID doesn't exist.
 */
const snmp_value_t *odb_find (const uint32_t *oid)
{
   return ((const snmp_value_t *) tree_find ((const tree_t *) odb,oid));
}

/*
 * Find the ObjectID in the ObjectId database that lexographically
 * succeeds the specified one or NULL if one doesn't exist. The
 * specified ObjectID doesn't need to exist in the database and can
 * also be a partial ObjectID. Check errno to see what error occurred.
 * The following errors is possible:
 *
 *       ENOMEM     Out of memory
 *       ENOENT     Specified ObjectID does not lexographically
 *                  preceed an existing ObjectID
 */
snmp_next_value_t *odb_find_next (const uint32_t *oid)
{
   branch_t *branch;
   snmp_next_value_t *next;

   if ((branch = tree_find_next ((const tree_t *) odb,oid)) == NULL)
	 return (NULL);

   if ((next = (snmp_next_value_t *) mem_alloc (sizeof (snmp_next_value_t))) == NULL)
	 {
		err0:
		mem_free (branch->branch);
		mem_free (branch);
		errno = ENOMEM;
		return (NULL);
	 }

   next->oid = branch->branch;
   memcpy (&next->value,branch->value,sizeof (snmp_value_t));

   if (next->value.type == BER_OCTET_STRING)
	 {
		uint8_t *tmp = next->value.data.OCTET_STRING.buf;

		if (next->value.data.OCTET_STRING.len)
		  {
			 if ((next->value.data.OCTET_STRING.buf = (uint8_t *) mem_alloc (next->value.data.OCTET_STRING.len)) == NULL)
			   {
				  err1:
				  mem_free (next);
				  goto err0;
			   }

			 memcpy (next->value.data.OCTET_STRING.buf,tmp,next->value.data.OCTET_STRING.len);
		  }
	 }
   else if (next->value.type == BER_OID)
	 {
		uint32_t *tmp = next->value.data.OID;

		if ((next->value.data.OID = (uint32_t *) mem_alloc ((tmp[0] + 1) * sizeof (uint32_t))) == NULL)
		  goto err1;

		memcpy (next->value.data.OID,tmp,(tmp[0] + 1) * sizeof (uint32_t));
	 }

   mem_free (branch);

   return (next);
}

/*
 * Remove an ObjectID subtree from the ObjectID database. Returns 0
 * if successful, or -1 if the ObjectID subtree doesn't exist.
 */
int odb_remove (const uint32_t *oid)
{
   return (tree_remove (&odb,oid,purge));
}

#define priv_printf(fmt,args...) log_printf_stub(filename,line,function,level,fmt,##args)
#define priv_puts(s) log_puts_stub(filename,line,function,level,s)
#define priv_putc(c) log_putc_stub(filename,line,function,level,c)

#define printable(x) (isprint(x) || isspace(x))

#define ADDR(x) (x) & 0x000000ff,((x) & 0x0000ff00) >> 8,((x) & 0x00ff0000) >> 16,((x) & 0xff000000) >> 24

static int show_stub (const char *filename,int line,const char *function,int level,const void *data)
{
   const snmp_value_t *value = (snmp_value_t *) data;

   switch (value->type)
	 {
	  case BER_INTEGER:
		return (priv_printf ("INTEGER %" PRId32,value->data.INTEGER));

	  case BER_Counter32:
		return (priv_printf ("Counter32 %" PRIu32,value->data.Counter32));

	  case BER_Gauge32:
		return (priv_printf ("Gauge32 %" PRIu32,value->data.Gauge32));

	  case BER_TimeTicks:
		  {
			 uint32_t day,ms,sec,min,hour;

			 if (priv_printf ("TimeTicks (%" PRIu32 ") ",value->data.TimeTicks) < 0)
			   return (-1);

			 ms = value->data.TimeTicks % 100, day = (value->data.TimeTicks - ms) / 100;
			 sec = day % 60, day = (day - sec) / 60;
			 min = day % 60, day = (day - min) / 60;
			 hour = day % 24, day = (day - hour) / 24;

			 if (day && priv_printf ("%" PRIu32 " day%s, ",day,day > 1 ? "s" : "") < 0)
			   return (-1);

			 return (priv_printf ("%.2" PRIu32 ":%.2" PRIu32 ":%.2" PRIu32 ".%.2" PRIu32,hour,min,sec,ms));
		  }

	  case BER_Counter64:
		return (priv_printf ("Counter64 %" PRIu64,value->data.Counter64));

	  case BER_OID:
		  {
			 int i;

			 if (priv_printf ("OBJECT IDENTIFIER %" PRIu32 ".%" PRIu32,value->data.OID[1] / 40,value->data.OID[1] % 40) < 0)
			   return (-1);

			 for (i = 2; i <= value->data.OID[0]; i++)
			   if (priv_printf (".%" PRIu32,value->data.OID[i]) < 0)
				 return (-1);

			 return (0);
		  }

	  case BER_OCTET_STRING:
		  {
			 uint32_t i;
			 int raw = 0;

			 if (priv_puts ("OCTET STRING") < 0)
			   return (-1);

			 for (i = 0; i < value->data.OCTET_STRING.len && !raw; i++)
			   if (!printable (value->data.OCTET_STRING.buf[i]))
				 raw = 1;

			 if (!raw)
			   {
				  if (priv_putc (' ') < 0)
					return (-1);

				  for (i = 0; i < value->data.OCTET_STRING.len; i++)
					if (priv_putc (value->data.OCTET_STRING.buf[i]) < 0)
					  return (-1);
			   }
			 else
			   {
				  for (i = 0; i < value->data.OCTET_STRING.len; i++)
					if (priv_printf (" %.2x",value->data.OCTET_STRING.buf[i]) < 0)
					  return (-1);
			   }

			 return (0);
		  }

	  case BER_IpAddress:
		return (priv_printf ("IpAddress %u.%u.%u.%u",ADDR(value->data.IpAddress)));

	  case BER_NULL:
		return (priv_printf ("No Such Name"));

	  default:
		return (priv_printf ("(unknown)"));
	 }
}

#undef priv_putc
#undef priv_puts
#undef priv_printf

/*
 * Display an ObjectID database. Returns 0 if successful, or -1 if some
 * error occurred. Check errno to see what error occurred.
 */
int odb_show_stub (const char *filename,int line,const char *function,int level)
{
   return (tree_show_stub (filename,line,function,level,(const tree_t *) odb,show_stub));
}

