#ifndef TREE_H
#define TREE_H

/*
 * Copyright (c) Abraham vd Merwe <abz@blio.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	  notice, this list of conditions and the following disclaimer in the
 *	  documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of other contributors
 *	  may be used to endorse or promote products derived from this software
 *	  without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>

#include <tinysnmp/compat/inttypes.h>

typedef struct
{
   uint32_t *branch;
   void *value;
} branch_t;

typedef struct tree_r
{
   enum { TREE_NODE, TREE_VALUE } type;
   union
	 {
		uint32_t node;
		void *value;
	 } data;
   struct tree_r *parent;
   struct tree_r *sibling;
   struct tree_r *child;
} tree_t;

typedef int (*show_func_t)(const char *filename,int line,const char *function,int level,const void *value);

/*
 * Destroy a tree. When this function returns, the tree is
 * in exactly the same state as when initialized.
 */
extern void tree_destroy (tree_t **tree,void (*purge)(void *));

/*
 * Add a new branch to the tree. Returns 0 if successful, -1
 * otherwise. Check errno to see what error occurred. The
 * following errors is possible:
 *
 *       ENOMEM     Out of memory
 *       EEXIST     A branch like this (or superset of) already exist
 */
extern int tree_add_stub (const char *filename,int line,const char *function,tree_t **tree,const branch_t *branch);

/*
 * Update the value of an existing branch. Returns the old value
 * of the branch if successful, or NULL of the branch doesn't
 * exist.
 */
extern void *tree_update (tree_t *tree,branch_t *branch);

/*
 * Find a branch in a tree. Return the value of an existing
 * branch in the tree if successful, or NULL if the branch
 * doesn't exist.
 */
extern const void *tree_find (const tree_t *tree,const uint32_t *branch);

/*
 * Find the branch in the tree that lexographically succeeds
 * the specified one or NULL if one doesn't exist. The specified
 * branch doesn't need to exist in the tree and can also be a
 * partial branch. Check errno to see what error occurred. The
 * following errors is possible:
 *
 *       ENOMEM     Out of memory
 *       ENOENT     Specified branch does not lexographically
 *                  preceed an existing branch
 */
extern branch_t *tree_find_next (const tree_t *tree,const uint32_t *branch);

/*
 * Remove a branch (or part thereof) from a tree. Returns 0 if
 * successful, or -1 if the branch subtree doesn't exist.
 */
extern int tree_remove (tree_t **tree,const uint32_t *branch,void (*purge)(void *));

/*
 * Display a tree. Returns 0 if successful, or -1 if some error
 * occurred. Check errno to see what error occurred.
 *
 * The show_stub function should return 0 if successful, or -1 if
 * some error occurred. If an error did occur, the show function
 * should set errno.
 */
extern int tree_show_stub (const char *filename,int line,const char *function,int level,const tree_t *tree,show_func_t show_stub);

#endif	/* #ifndef TREE_H */
