
/*
 * Copyright (c) Abraham vd Merwe <abz@blio.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	  notice, this list of conditions and the following disclaimer in the
 *	  documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of other contributors
 *	  may be used to endorse or promote products derived from this software
 *	  without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* we need this for vsscanf() */
#define _ISOC99_SOURCE

#include <stdarg.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

#include <debug/log.h>
#include <debug/memory.h>

#include <tinysnmp/tinysnmp.h>
#include <tinysnmp/agent/odb.h>
#include <tinysnmp/agent/mod.h>
#include <tinysnmp/compat/inttypes.h>

#include <abz/getline.h>
#include <abz/trim.h>

#define _PROC_MEMINFO	"/proc/meminfo"

#ifndef SCNu64
#define SCNu64 "llu"
#endif	/* #ifndef SCNu64 */

typedef enum
{
   TYPE_MEMORY,
   TYPE_SWAP
} mem_type_t;

typedef struct
{
   uint32_t mem_total;
   uint32_t mem_free;
   uint32_t mem_buffer;
   uint32_t mem_cache;
   uint32_t swap_total;
   uint32_t swap_free;
} mem_t;

static int find (int fd,int args,const char *fmt, ...)
{
   va_list ap;
   char *buf;

   va_start (ap,fmt);

   while ((buf = getline (fd)) != NULL)
	 {
		trim (buf);
		if (args == vsscanf (buf,fmt,ap))
		  {
			 mem_free (buf);
			 return (0);
		  }
		mem_free (buf);
	 }

   va_end (ap);

   return (-1);
}

static int meminfo (const struct mod_r *mod,mem_t *mem,mem_type_t type)
{
   int fd,result;
   uint64_t m_total,m_used,m_free,m_shared,m_buffer,m_cached;

   if ((fd = open (_PROC_MEMINFO,O_RDONLY)) < 0)
	 {
		log_printf (LOG_ERROR,"%s: open %s failed: %s\n",mod->name,_PROC_MEMINFO,strerror (errno));
		return (-1);
	 }

   if (type == TYPE_MEMORY)
	 result = find (fd,6,
					"Mem: %" SCNu64 " %" SCNu64 " %" SCNu64 " %" SCNu64 " %" SCNu64 " %" SCNu64 "\n",
					&m_total,&m_used,&m_free,&m_shared,&m_buffer,&m_cached);
   else
	 result = find (fd,3,"Swap: %" SCNu64 " %" SCNu64 " %" SCNu64 "\n",&m_total,&m_used,&m_free);

   close (fd);

   if (result < 0)
	 {
		log_printf (LOG_ERROR,"%s: input/output error or unsupported %s\n",mod->name,_PROC_MEMINFO);
		return (-1);
	 }

   if (type == TYPE_MEMORY)
	 {
		mem->mem_total = m_total / 1024;
		mem->mem_free = m_free / 1024;
		mem->mem_buffer = m_buffer / 1024;
		mem->mem_cache = m_cached / 1024;
	 }
   else
	 {
		mem->swap_total = m_total / 1024;
		mem->swap_free = m_free / 1024;
	 }

   return (0);
}

static int memory_update (const struct mod_r *mod)
{
   /* iso.org.dod.internet.private.enterprises.frogfoot.servers.system.resources.memory.xx.0 */
   uint32_t oid[13] = { 12, 43, 6, 1, 4, 1, 10002, 1, 1, 1, 1, 0, 0 };
   snmp_value_t value;
   mem_t mem;

   if (meminfo (mod,&mem,TYPE_MEMORY) < 0) return (-1);

   odb_remove (mod->oid);

   value.type = BER_Gauge32;

   oid[11] = 1;
   value.data.Gauge32 = mem.mem_total;

   if (odb_add (oid,&value) < 0)
	 {
		log_printf (LOG_ERROR,"%s: couldn't add servers.system.resources.memory.memTotal.0: %s\n",mod->name,strerror (errno));
		return (-1);
	 }

   oid[11] = 2;
   value.data.Gauge32 = mem.mem_free;

   if (odb_add (oid,&value) < 0)
	 {
		log_printf (LOG_ERROR,"%s: couldn't add servers.system.resources.memory.memFree.0: %s\n",mod->name,strerror (errno));
		odb_remove (mod->oid);
		return (-1);
	 }

   oid[11] = 3;
   value.data.Gauge32 = mem.mem_buffer;

   if (odb_add (oid,&value) < 0)
	 {
		log_printf (LOG_ERROR,"%s: couldn't add servers.system.resources.memory.memBuffer.0: %s\n",mod->name,strerror (errno));
		odb_remove (mod->oid);
		return (-1);
	 }

   oid[11] = 4;
   value.data.Gauge32 = mem.mem_cache;

   if (odb_add (oid,&value) < 0)
	 {
		log_printf (LOG_ERROR,"%s: couldn't add servers.system.resources.memory.memCache.0: %s\n",mod->name,strerror (errno));
		odb_remove (mod->oid);
		return (-1);
	 }

   return (0);
}

static int swap_update (const struct mod_r *mod)
{
   /* iso.org.dod.internet.private.enterprises.frogfoot.servers.system.resources.swap.xx.0 */
   uint32_t oid[13] = { 12, 43, 6, 1, 4, 1, 10002, 1, 1, 1, 2, 0, 0 };
   snmp_value_t value;
   mem_t swap;

   if (meminfo (mod,&swap,TYPE_SWAP) < 0) return (-1);

   odb_remove (mod->oid);

   value.type = BER_Gauge32;

   oid[11] = 1;
   value.data.Gauge32 = swap.swap_total;

   if (odb_add (oid,&value) < 0)
	 {
		log_printf (LOG_ERROR,"%s: couldn't add servers.system.resources.swap.swapTotal.0: %s\n",mod->name,strerror (errno));
		return (-1);
	 }

   oid[11] = 2;
   value.data.Gauge32 = swap.swap_free;

   if (odb_add (oid,&value) < 0)
	 {
		log_printf (LOG_ERROR,"%s: couldn't add servers.system.resources.swap.swapFree.0: %s\n",mod->name,strerror (errno));
		odb_remove (mod->oid);
		return (-1);
	 }

   return (0);
}

static int storage_update (const struct mod_r *mod)
{
   return (0);
}

/* iso.org.dod.internet.private.enterprises.frogfoot.servers.system.resources.memory */
static const uint32_t memory_oid[11] = { 10, 43, 6, 1, 4, 1, 10002, 1, 1, 1, 1 };

/* iso.org.dod.internet.private.enterprises.frogfoot.servers.system.resources.swap */
static const uint32_t swap_oid[11] = { 10, 43, 6, 1, 4, 1, 10002, 1, 1, 1, 2 };

/* iso.org.dod.internet.private.enterprises.frogfoot.servers.system.resources.storage */
static const uint32_t storage_oid[11] = { 10, 43, 6, 1, 4, 1, 10002, 1, 1, 1, 3 };

const mod_t mod[] =
{
   { "frogfoot/memory", NULL, memory_update, NULL, memory_oid },
   { "frogfoot/swap", NULL, swap_update, NULL, swap_oid },
   { "frogfoot/storage", NULL, storage_update, NULL, storage_oid },
   { NULL, NULL, NULL, NULL, NULL }
};

