
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <time.h>

#include <tinysnmp/compat/inttypes.h>

#include <abz/atou32.h>
#include <abz/contigmask.h>

#include "ipc.h"

#include <debug/log.h>
#include <debug/memory.h>

#define ADDR(x) (x) & 0x000000ff,((x) & 0x0000ff00) >> 8,((x) & 0x00ff0000) >> 16,((x) & 0xff000000) >> 24

#define QUALITY_BRONZE(x,y) ((x) == IPC_QUALITY_BRONZE ? "bronze" : (y))
#define QUALITY_SILVER(x,y) ((x) == IPC_QUALITY_SILVER ? "silver" : (y))
#define QUALITY_GOLD(x,y) ((x) == IPC_QUALITY_GOLD ? "gold" : (y))
#define QUALITY_PLATINUM(x,y) ((x) == IPC_QUALITY_PLATINUM ? "platinum" : (y))
#define QUALITY(x) (QUALITY_BRONZE (x,QUALITY_SILVER (x,QUALITY_GOLD (x,QUALITY_PLATINUM (x,"unknown")))))

static void print_client_info (int level,const ipc_client_info_t *info)
{
   uint32_t i;

   log_printf (level,
			   "id ......... %u\n"
			   "name ....... %s\n"
			   "quality .... %s\n"
			   "input ...... %" PRIu32 " Kbits\n"
			   "output ..... %" PRIu32 " Kbits\n"
			   "networks ...",
			   info->id,
			   info->name,
			   QUALITY (info->quality),
			   info->input / 1000,
			   info->output / 1000);

   for (i = 0; i < info->n; i++)
	 {
		uint8_t mask;

		log_printf (level," %" PRIu32 ".%" PRIu32 ".%" PRIu32 ".%" PRIu32 "/",ADDR (info->network[i].address));

		if ((mask = contigmask (info->network[i].netmask)) < 0xff)
		  log_printf (level,"%" PRIu32,mask);
		else
		  log_printf (level,"%" PRIu32 ".%" PRIu32 ".%" PRIu32 ".%" PRIu32,ADDR (info->network[i].netmask));
	 }

   log_putc (level,'\n');
}

static void print_client_counters (int level,const ipc_client_counters_t *counters)
{
   uint32_t i;
   uint64_t recv[IPC_PROTO_TOTAL + 1],xmit[IPC_PROTO_TOTAL];
   uint8_t mask[IPC_PROTO_TOTAL + 1];
   static const char *prefix[IPC_PROTO_TOTAL + 1] =
	 {
		[IPC_PROTO_UNKNOWN] "unknown ...",
		[IPC_PROTO_FTP    ] "ftp .......",
		[IPC_PROTO_SSH    ] "ssh .......",
		[IPC_PROTO_TELNET ] "telnet ....",
		[IPC_PROTO_SMTP   ] "smtp ......",
		[IPC_PROTO_DNS    ] "dns .......",
		[IPC_PROTO_TFTP   ] "tftp ......",
		[IPC_PROTO_HTTP   ] "http ......",
		[IPC_PROTO_POP3   ] "pop3 ......",
		[IPC_PROTO_IMAP   ] "imap ......",
		[IPC_PROTO_SNMP   ] "snmp ......",
		[IPC_PROTO_IRC    ] "irc .......",
		[IPC_PROTO_TOTAL  ] "total ....."
	 };

   for (i = 0; i <= IPC_PROTO_TOTAL; i++) mask[i] = 0;

   for (i = 0; i < counters->n; i++)
	 {
		mask[counters->counter[i].protocol] |= 1 << counters->counter[i].type;
		if (counters->counter[i].type == IPC_TYPE_RECEIVED)
		  recv[counters->counter[i].protocol] = counters->counter[i].value;
		else
		  xmit[counters->counter[i].protocol] = counters->counter[i].value;
	 }

   for (i = 0; i <= IPC_PROTO_TOTAL; i++)
	 if (mask[i])
	   {
		  static char a[64],b[64];

		  sprintf (a,"%" PRIu64,recv[i]);
		  sprintf (b,"%" PRIu64,xmit[i]);

		  log_printf (level,"%s %7s bytes in %15s bytes out\n",
					  prefix[i],
					  mask[i] & (1 << IPC_TYPE_RECEIVED) ? a : "UN",
					  mask[i] & (1 << IPC_TYPE_TRANSMITTED) ? b : "UN");
	   }
}

static void fail (const char *fmt, ...)
{
   va_list ap;
   va_start (ap,fmt);
   if (fmt != NULL) log_vprintf (LOG_ERROR,fmt,ap);
   va_end (ap);
   exit (EXIT_FAILURE);
}

static void cleanup ()
{
   mem_close ();
   log_close ();
}

int main (int argc,char *argv[])
{
   ipc_client_t clients;
   uint32_t i;
   time_t timestamp;

   mem_open (fail);
   log_open (NULL,LOG_NOISY,LOG_HAVE_COLORS);
   atexit (cleanup);

   if (argc != 2)
	 {
		const char *progname;
		(progname = strrchr (argv[0],'/')) ? progname++ : (progname = argv[0]);
		fail ("usage: %s <socket>\n"
			  "       %s -h | --help\n",
			  progname,progname);
	 }

   if (ipc_get_clients (argv[1],&clients) < 0)
	 exit (EXIT_FAILURE);

   timestamp = clients.uptime;
   log_printf (LOG_NORMAL,"last updated: %s\n",ctime (&timestamp));

   for (i = 0; i < clients.n; i++)
	 {
		log_printf (LOG_NORMAL,
					"client %" PRIu32 "\n"
					"==================\n\n",
					i);
		print_client_info (LOG_NORMAL,clients.info + i);
		log_putc (LOG_NORMAL,'\n');
		print_client_counters (LOG_NORMAL,clients.counter + i);
		log_putc (LOG_NORMAL,'\n');
	 }

   ipc_destroy_clients (&clients);

   exit (EXIT_SUCCESS);
}

