
/*
 * Copyright (c) Abraham vd Merwe <abz@blio.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	  notice, this list of conditions and the following disclaimer in the
 *	  documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of other contributors
 *	  may be used to endorse or promote products derived from this software
 *	  without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <errno.h>
#include <unistd.h>
#include <string.h>

#include <debug/log.h>

#include <tinysnmp/compat/inttypes.h>

#include "ipc.h"

typedef ssize_t (*method_t) (int,void *,size_t);

#define METHOD(x) { #x, (method_t) x }

typedef struct
{
   const char *name;
   method_t method;
} transfer_t;

static const transfer_t methods[] = { [IPC_TRANSFER_READ] METHOD(read), [IPC_TRANSFER_WRITE] METHOD(write) };

/*
 * transfer data between client and server.
 *
 *        level     log level
 *        fd        file descriptor of socket
 *        buf       pointer to buffer
 *        count     number of bytes to transfer
 *        type      transfer type (read/write)
 *
 * returns 0 if successful, -1 if some error occurred. errors are
 * logged automatically.
 */
int ipc_transfer_stub (const char *function,int level,int fd,void *buf,ssize_t count,ipc_transfer_t type)
{
   ssize_t result;
   uint8_t *s = (uint8_t *) buf;

   do
	 {
		result = methods[type].method (fd,s,count);
		if (result > 0) count -= result, s += result;
	 }
   while ((result > 0 && count > 0) || (result < 0 && errno == EINTR));

   if (result < 0)
	 {
		log_printf (level,"%s: %s failed: %s\n",function,methods[type].name,strerror (errno));
		return (-1);
	 }

   if (!result)
	 {
		log_printf (level,"%s: short %s count on socket. remote side closed connection?\n",function,methods[type].name);
		return (-1);
	 }

   if (count < 0)
	 {
		log_printf (level,"%s: %s returned bogus result. bug in socket implementation?\n",function,methods[type].name);
		return (-1);
	 }

   return (0);
}

