
/*
 * Copyright (c) Abraham vd Merwe <abz@blio.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	  notice, this list of conditions and the following disclaimer in the
 *	  documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of other contributors
 *	  may be used to endorse or promote products derived from this software
 *	  without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/un.h>

#include <debug/log.h>
#include <debug/memory.h>

#include <tinysnmp/compat/inttypes.h>

#include <abz/aton.h>

#include "ipc.h"

#define LEVEL LOG_ERROR

#define invalid_packet() invalid_packet_stub(__FILE__,__LINE__,__FUNCTION__)
static void invalid_packet_stub (const char *filename,int line,const char *function)
{
   log_printf_stub (filename,line,function,LEVEL,"%s: invalid packet format. broken server?\n",function);
}

#define out_of_memory() out_of_memory_stub(__FILE__,__LINE__,__FUNCTION__)
static void out_of_memory_stub (const char *filename,int line,const char *function)
{
   log_printf_stub (filename,line,function,LEVEL,"%s: unable to allocate memory for temporary storage: %s\n",function,strerror (errno));
}

static int ipc_open (const char *s)
{
   struct sockaddr_un addr;
   int fd;

   if ((fd = socket (PF_UNIX,SOCK_STREAM,0)) < 0)
	 {
		log_printf (LEVEL,"unable to create socket: %s\n",strerror (errno));
		return (-1);
	 }

   addr.sun_family = AF_UNIX;
   strncpy (addr.sun_path,s,sizeof (addr.sun_path));
   addr.sun_path[sizeof (addr.sun_path) - 1] = '\0';

   if (connect (fd,(struct sockaddr *) &addr,SUN_LEN (&addr)) < 0)
	 {
		log_printf (LEVEL,"unable to connect to socket %s: %s\n",s,strerror (errno));
		close (fd);
		return (-1);
	 }

   return (fd);
}

static int put (int fd,uint32_t cmd,uint32_t client)
{
   uint32_t buf[3] = { 0, cmd, client };
   uint32_t size = sizeof (buf);

   if (cmd == IPC_PACKET_UPDATE || cmd == IPC_PACKET_CLIENTS || cmd == IPC_PACKET_UPTIME)
	 size -= 4;

   buf[0] = size;

   return (ipc_transfer (LEVEL,fd,buf,size,IPC_TRANSFER_WRITE));
}

static int get_clients (int fd,uint32_t *clients)
{
   uint32_t buf[3];

   if (ipc_transfer (LEVEL,fd,buf,sizeof (buf),IPC_TRANSFER_READ) < 0)
	 return (-1);

   if (buf[0] != 12 || buf[1] != IPC_PACKET_CLIENTS)
	 {
		invalid_packet ();
		return (-1);
	 }

   *clients = buf[2];

   return (0);
}

static int decode_64 (uint64_t *value,uint8_t **s,uint32_t *len)
{
   if (*len < 8)
	 {
		invalid_packet ();
		return (-1);
	 }

   *value = *((uint64_t *) *s), *len -= 8, (*s) += 8;

   return (0);
}

static int decode_32 (uint32_t *value,uint8_t **s,uint32_t *len)
{
   if (*len < 4)
	 {
		invalid_packet ();
		return (-1);
	 }

   *value = *((uint32_t *) *s), *len -= 4, (*s) += 4;

   return (0);
}

static int decode_8 (uint8_t *value,uint8_t **s,uint32_t *len)
{
   if (!*len)
	 {
		invalid_packet ();
		return (-1);
	 }

   *value = *(*s)++, (*len)--;

   return (0);
}

static char *decode_name (uint8_t **s,uint32_t *len)
{
   uint32_t i,length;
   char *result;

   if (decode_32 (&length,s,len) < 0)
	 return (NULL);

   if ((result = (char *) mem_alloc (sizeof (char) * (length + 1))) == NULL)
	 {
		out_of_memory ();
		return (NULL);
	 }

   for (i = 0; i < length; i++)
	 if (decode_8 (result + i,s,len) < 0)
	   {
		  mem_free (result);
		  return (NULL);
	   }

   result[i] = '\0';

   return (result);
}

static int decode_quality (uint8_t *quality,uint8_t **s,uint32_t *len)
{
   if (decode_8 (quality,s,len) < 0)
	 return (-1);

   if (*quality > IPC_QUALITY_PLATINUM)
	 {
		invalid_packet ();
		return (-1);
	 }

   return (0);
}

static int decode_network (network_t *network,uint8_t **s,uint32_t *len)
{
   if (decode_32 (&network->address,s,len) < 0 || decode_32 (&network->netmask,s,len) < 0)
	 return (-1);

   return (0);
}

static int get_info (int fd,ipc_client_info_t *info)
{
   uint32_t i,len = 0,hdr[2];
   uint8_t *s,*buf = NULL;
   int result = -1;

   memset (info,0L,sizeof (ipc_client_info_t));

   do
	 {
		if (ipc_transfer (LEVEL,fd,hdr,sizeof (hdr),IPC_TRANSFER_READ) < 0)
		  break;

		if (hdr[0] < 29 || hdr[1] != IPC_PACKET_INFO)
		  {
			 invalid_packet ();
			 break;
		  }

		len = hdr[0] - 8;

		if ((buf = (uint8_t *) mem_alloc (len)) == NULL)
		  {
			 out_of_memory ();
			 break;
		  }

		s = buf;

		if (ipc_transfer (LEVEL,fd,buf,len,IPC_TRANSFER_READ) < 0 ||
			decode_32 (&info->id,&s,&len) < 0 ||
			(info->name = decode_name (&s,&len)) == NULL ||
			decode_quality (&info->quality,&s,&len) < 0 ||
			decode_32 (&info->input,&s,&len) < 0 ||
			decode_32 (&info->output,&s,&len) < 0 ||
			decode_32 (&info->n,&s,&len) < 0)
		  break;

		if ((info->network = (network_t *) mem_alloc (info->n * sizeof (network_t))) == NULL)
		  break;

		result = 0;

		for (i = 0; !result && i < info->n; i++)
		  result = decode_network (info->network + i,&s,&len);
	 }
   while (0);

   if (buf != NULL) mem_free (buf);

   if (len)
	 {
		invalid_packet ();
		result = -1;
	 }

   if (result < 0)
	 {
		if (info->name != NULL) mem_free (info->name);
		if (info->network != NULL) mem_free (info->network);
		return (-1);
	 }

   return (0);
}

static int decode_counter (ipc_counter_t *counter,uint8_t **s,uint32_t *len)
{
   if (decode_8 (&counter->protocol,s,len) < 0 || decode_8 (&counter->type,s,len) < 0 || decode_64 (&counter->value,s,len) < 0)
	 return (-1);

   if (counter->protocol > IPC_PROTO_TOTAL || counter->type > IPC_TYPE_TRANSMITTED)
	 {
		invalid_packet ();
		return (-1);
	 }

   return (0);
}

static int get_counters (int fd,ipc_client_counters_t *counters)
{
   uint32_t i,len = 0,hdr[3];
   uint8_t *s,*buf = NULL;
   int result = -1;

   memset (counters,0L,sizeof (ipc_client_counters_t));

   do
	 {
		if (ipc_transfer (LEVEL,fd,hdr,sizeof (hdr),IPC_TRANSFER_READ) < 0)
		  break;

		if (hdr[0] < sizeof (hdr) || hdr[0] != sizeof (hdr) + hdr[2] * 10 || hdr[1] != IPC_PACKET_COUNTERS)
		  {
			 invalid_packet ();
			 break;
		  }

		counters->n = hdr[2];
		len = counters->n * 10;

		if (!counters->n)
		  {
			 result = 0;
			 break;
		  }

		if ((buf = (uint8_t *) mem_alloc (len)) == NULL)
		  {
			 out_of_memory ();
			 break;
		  }

		s = buf;

		if (ipc_transfer (LEVEL,fd,buf,len,IPC_TRANSFER_READ) < 0)
		  break;

		if ((counters->counter = (ipc_counter_t *) mem_alloc (sizeof (ipc_counter_t) * counters->n)) == NULL)
		  {
			 out_of_memory ();
			 break;
		  }

		result = 0;

		for (i = 0; !result && i < counters->n; i++)
		  result = decode_counter (counters->counter + i,&s,&len);
	 }
   while (0);

   if (buf != NULL) mem_free (buf);

   if (len)
	 {
		invalid_packet ();
		result = -1;
	 }

   if (result < 0)
	 {
		if (counters->counter != NULL) mem_free (counters->counter);
		return (-1);
	 }

   return (0);
}

static int get_uptime (int fd,uint64_t *timestamp)
{
   uint32_t buf[4];

   if (ipc_transfer (LEVEL,fd,buf,sizeof (buf),IPC_TRANSFER_READ) < 0)
	 return (-1);

   if (buf[0] != sizeof (buf) || buf[1] != IPC_PACKET_UPTIME)
	 return (-1);

   *timestamp = ((uint64_t *) buf)[1];

   return (0);
}

/*
 * Retrieve a client structure from the trafficd server. Returns
 * 0 if successful, -1 if some error occurred. Errors are logged
 * automatically.
 */
int ipc_get_clients (const char *socket,ipc_client_t *clients)
{
   int fd,result = -1,err = 0;
   uint32_t i,j;

   memset (clients,0L,sizeof (ipc_client_t));

   do
	 {
		if ((fd = ipc_open (socket)) < 0 || put (fd,IPC_PACKET_UPDATE,0) < 0)
		  break;

		close (fd);

		if ((fd = ipc_open (socket)) < 0 || put (fd,IPC_PACKET_UPTIME,0) < 0 || get_uptime (fd,&clients->uptime) < 0)
		  break;

		close (fd);

		if ((fd = ipc_open (socket)) < 0 || put (fd,IPC_PACKET_CLIENTS,0) < 0 || get_clients (fd,&clients->n) < 0)
		  break;

		close (fd);
		fd = -1;

		if (!clients->n)
		  {
			 result = 0;
			 break;
		  }

		if ((clients->info = (ipc_client_info_t *) mem_alloc (clients->n * sizeof (ipc_client_info_t))) == NULL)
		  {
			 out_of_memory ();
			 break;
		  }

		for (i = 0; i < clients->n; i++)
		  {
			 if ((fd = ipc_open (socket)) < 0 || put (fd,IPC_PACKET_INFO,i) < 0 || get_info (fd,clients->info + i) < 0)
			   {
				  for (j = 0; j < i; j++)
					{
					   mem_free (clients->info[j].name);
					   mem_free (clients->info[j].network);
					}
				  err = 1;
				  break;
			   }

			 close (fd);
			 fd = -1;
		  }

		if (err) break;

		if ((clients->counter = (ipc_client_counters_t *) mem_alloc (clients->n * sizeof (ipc_client_counters_t))) == NULL)
		  break;

		for (i = 0; i < clients->n; i++)
		  {
			 if ((fd = ipc_open (socket)) < 0 || put (fd,IPC_PACKET_COUNTERS,i) < 0 || get_counters (fd,clients->counter + i) < 0)
			   {
				  for (j = 0; j < i; j++)
					if (clients->counter[j].counter != NULL)
					  mem_free (clients->counter[j].counter);
				  err = 1;
				  break;
			   }

			 close (fd);
			 fd = -1;
		  }

		if (!err) result = 0;
	 }
   while (0);

   if (fd > 0) close (fd);

   if (result < 0)
	 {
		if (clients->info != NULL) mem_free (clients->info);
		if (clients->counter != NULL) mem_free (clients->counter);
	 }

   return (result);
}

/*
 * Free memory allocated by the ipc_get_client() structure.
 */
void ipc_destroy_clients (ipc_client_t *clients)
{
   uint32_t i;

   for (i = 0; i < clients->n; i++)
	 {
		mem_free (clients->info[i].name);
		mem_free (clients->info[i].network);
		if (clients->counter[i].counter != NULL)
		  mem_free (clients->counter[i].counter);
	 }

   if (clients->info != NULL) mem_free (clients->info);
   if (clients->counter != NULL) mem_free (clients->counter);
}

