
/*
 * Copyright (c) Abraham vd Merwe <abz@blio.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	  notice, this list of conditions and the following disclaimer in the
 *	  documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of other contributors
 *	  may be used to endorse or promote products derived from this software
 *	  without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <errno.h>
#include <time.h>
#include <string.h>

#include <debug/log.h>
#include <debug/memory.h>

#include <tinysnmp/tinysnmp.h>
#include <tinysnmp/agent/odb.h>
#include <tinysnmp/agent/mod.h>
#include <tinysnmp/compat/inttypes.h>

#include <ber/ber.h>

#include "ipc.h"
#include "config.h"

static const config_t *config;

typedef struct
{
   uint32_t type;
   uint8_t mask;
   uint64_t recv;
   uint64_t xmit;
} priv_proto_t;

static int add_uptime (const struct mod_r *mod,uint64_t uptime)
{
   /* iso.org.dod.internet.private.enterprises.frogfoot.traffic.clientUpTime.0 */
   static const uint32_t clientUpTime[10] = { 9, 43, 6, 1, 4, 1, 10002, 2, 1, 0 };
   snmp_value_t value;

   value.type = BER_TimeTicks;
   value.data.TimeTicks = difftime (time (NULL),uptime) * 10;

   if (odb_add (clientUpTime,&value) < 0)
	 {
		log_printf (LOG_ERROR,"%s: couldn't add traffic.clientUpTime.0: %s\n",mod->name,strerror (errno));
		return (-1);
	 }

   return (0);
}

static int add_client_number (const struct mod_r *mod,uint32_t n)
{
   /* iso.org.dod.internet.private.enterprises.frogfoot.traffic.clientNumber.0 */
   static const uint32_t clientNumber[10] = { 9, 43, 6, 1, 4, 1, 10002, 2, 2, 0 };
   snmp_value_t value;

   value.type = BER_INTEGER;
   value.data.INTEGER = n;

   if (odb_add (clientNumber,&value) < 0)
	 {
		log_printf (LOG_ERROR,"%s: couldn't add traffic.clientNumber.0: %s\n",mod->name,strerror (errno));
		return (-1);
	 }

   return (0);
}

static int add_client_network (const struct mod_r *mod,uint32_t index,network_t *network,uint32_t n)
{
   /* iso.org.dod.internet.private.enterprises.frogfoot.traffic.networkTable.networkEntry.xx.yy.zz */
   static uint32_t networkTable[13] = { 12, 43, 6, 1, 4, 1, 10002, 2, 5, 1, 0, 0, 0 };
   snmp_value_t value;
   static const char prefix[] = "couldn't add traffic.networkTable.networkEntry";

   networkTable[10] = 1;
   networkTable[11] = index;
   networkTable[12] = n;

   value.type = BER_INTEGER;
   value.data.INTEGER = n;

   if (odb_add (networkTable,&value) < 0)
	 {
		log_printf (LOG_ERROR,"%s: %s.netIndex.%" PRIu32 ".%" PRIu32 ": %s\n",mod->name,prefix,index,n,strerror (errno));
		return (-1);
	 }

   networkTable[10] = 2;
   value.type = BER_IpAddress;
   value.data.IpAddress = network->address;

   if (odb_add (networkTable,&value) < 0)
	 {
		log_printf (LOG_ERROR,"%s: %s.netAddress.%" PRIu32 ".%" PRIu32 ": %s\n",mod->name,prefix,index,n,strerror (errno));
		return (-1);
	 }

   networkTable[10] = 3;
   value.data.IpAddress = network->netmask;

   if (odb_add (networkTable,&value) < 0)
	 {
		log_printf (LOG_ERROR,"%s: %s.netMask.%" PRIu32 ".%" PRIu32 ": %s\n",mod->name,prefix,index,n,strerror (errno));
		return (-1);
	 }

   return (0);
}

static int add_client_protocol (const struct mod_r *mod,uint32_t index,priv_proto_t *proto,uint32_t n)
{
   /* iso.org.dod.internet.private.enterprises.frogfoot.traffic.protocolTable.protocolEntry.xx.yy.zz */
   static uint32_t protocolTable[13] = { 12, 43, 6, 1, 4, 1, 10002, 2, 4, 1, 0, 0, 0 };
   snmp_value_t value;
   static const char prefix[] = "couldn't add traffic.protocolTable.protocolEntry";

   protocolTable[10] = 1;
   protocolTable[11] = index;
   protocolTable[12] = n;

   value.type = BER_INTEGER;
   value.data.INTEGER = n;

   if (odb_add (protocolTable,&value) < 0)
	 {
		log_printf (LOG_ERROR,"%s: %s.protoIndex.%" PRIu32 ".%" PRIu32 ": %s\n",mod->name,prefix,index,n,strerror (errno));
		return (-1);
	 }

   protocolTable[10] = 2;
   value.data.INTEGER = proto->type;

   if (odb_add (protocolTable,&value) < 0)
	 {
		log_printf (LOG_ERROR,"%s: %s.protoType.%" PRIu32 ".%" PRIu32 ": %s\n",mod->name,prefix,index,n,strerror (errno));
		return (-1);
	 }

   value.type = BER_Counter64;

   if (proto->mask & (1 << IPC_TYPE_RECEIVED))
	 {
		protocolTable[10] = 3;
		value.data.Counter64 = proto->recv;

		if (odb_add (protocolTable,&value) < 0)
		  {
			 log_printf (LOG_ERROR,"%s: %s.protoReceived.%" PRIu32 ": %s\n",mod->name,prefix,n,strerror (errno));
			 return (-1);
		  }
	 }

   if (proto->mask & (1 << IPC_TYPE_TRANSMITTED))
	 {
		protocolTable[10] = 4;
		value.data.Counter64 = proto->xmit;

		if (odb_add (protocolTable,&value) < 0)
		  {
			 log_printf (LOG_ERROR,"%s: %s.protoSent.%" PRIu32 ": %s\n",mod->name,prefix,n,strerror (errno));
			 return (-1);
		  }
	 }

   return (0);
}

static int add_client (const struct mod_r *mod,const ipc_client_info_t *info,const ipc_client_counters_t *counters,uint32_t n)
{
   /* iso.org.dod.internet.private.enterprises.frogfoot.traffic.clientTable.clientEntry.xx.yy */
   static uint32_t clientTable[12] = { 11, 43, 6, 1, 4, 1, 10002, 2, 3, 1, 0, 0 };
   snmp_value_t value;
   uint32_t i;
   static const char prefix[] = "couldn't add traffic.clientTable.clientEntry";
   uint64_t recv[IPC_PROTO_TOTAL + 1],xmit[IPC_PROTO_TOTAL];
   uint8_t mask[IPC_PROTO_TOTAL + 1];
   priv_proto_t proto;

   clientTable[10] = 1;
   clientTable[11] = info->id;

   value.type = BER_INTEGER;
   value.data.INTEGER = n;

   if (odb_add (clientTable,&value) < 0)
	 {
		log_printf (LOG_ERROR,"%s: %s.clientIndex.%" PRIu32 ": %s\n",mod->name,prefix,n,strerror (errno));
		return (-1);
	 }

   clientTable[10] = 2;

   value.type = BER_OCTET_STRING;
   value.data.OCTET_STRING.len = strlen (info->name);
   value.data.OCTET_STRING.buf = (uint8_t *) info->name;

   if (odb_add (clientTable,&value) < 0)
	 {
		log_printf (LOG_ERROR,"%s: %s.clientName.%" PRIu32 ": %s\n",mod->name,prefix,n,strerror (errno));
		return (-1);
	 }

   clientTable[10] = 3;

   value.type = BER_INTEGER;
   value.data.INTEGER = 1;

   switch (info->quality)
	 {
	  case IPC_QUALITY_PLATINUM:	value.data.INTEGER++;
	  case IPC_QUALITY_GOLD:		value.data.INTEGER++;
	  case IPC_QUALITY_SILVER:		value.data.INTEGER++;
		/* IPC_QUALITY_BRONZE is the default */
	 }

   if (odb_add (clientTable,&value) < 0)
	 {
		log_printf (LOG_ERROR,"%s: %s.serviceQuality.%" PRIu32 ": %s\n",mod->name,prefix,n,strerror (errno));
		return (-1);
	 }

   clientTable[10] = 4;

   value.type = BER_Gauge32;
   value.data.Gauge32 = info->input;

   if (odb_add (clientTable,&value) < 0)
	 {
		log_printf (LOG_ERROR,"%s: %s.inputRate.%" PRIu32 ": %s\n",mod->name,prefix,n,strerror (errno));
		return (-1);
	 }

   clientTable[10] = 5;

   value.data.Gauge32 = info->output;

   if (odb_add (clientTable,&value) < 0)
	 {
		log_printf (LOG_ERROR,"%s: %s.outputRate.%" PRIu32 ": %s\n",mod->name,prefix,n,strerror (errno));
		return (-1);
	 }

   value.type = BER_Counter64;

   for (i = 0; i <= IPC_PROTO_TOTAL; i++) mask[i] = 0;
   for (i = 0; i < counters->n; i++)
	 {
		mask[counters->counter[i].protocol] |= 1 << counters->counter[i].type;
		if (counters->counter[i].type == IPC_TYPE_RECEIVED)
		  recv[counters->counter[i].protocol] = counters->counter[i].value;
		else
		  xmit[counters->counter[i].protocol] = counters->counter[i].value;
	 }

   if (mask[IPC_PROTO_TOTAL] & (1 << IPC_TYPE_RECEIVED))
	 {
		clientTable[10] = 6;
		value.data.Counter64 = recv[IPC_PROTO_TOTAL];

		if (odb_add (clientTable,&value) < 0)
		  {
			 log_printf (LOG_ERROR,"%s: %s.bytesReceived.%" PRIu32 ": %s\n",mod->name,prefix,n,strerror (errno));
			 return (-1);
		  }
	 }

   if (mask[IPC_PROTO_TOTAL] & (1 << IPC_TYPE_TRANSMITTED))
	 {
		clientTable[10] = 7;
		value.data.Counter64 = xmit[IPC_PROTO_TOTAL];

		if (odb_add (clientTable,&value) < 0)
		  {
			 log_printf (LOG_ERROR,"%s: %s.bytesSent.%" PRIu32 ": %s\n",mod->name,prefix,n,strerror (errno));
			 return (-1);
		  }
	 }

   for (i = 0; i < info->n; i++)
	 if (add_client_network (mod,n,info->network + i,i + 1) < 0)
	   return (-1);

   for (i = 0; i < IPC_PROTO_TOTAL; i++)
	 if (mask[i])
	   {
		  proto.type = i;
		  proto.mask = mask[i];
		  proto.recv = recv[i];
		  proto.xmit = xmit[i];

		  if (add_client_protocol (mod,n,&proto,i + 1) < 0)
			return (-1);
	   }

   return (0);
}

static int traffic_update (const struct mod_r *mod)
{
   ipc_client_t clients;
   int result = -1;
   uint32_t i;

   odb_remove (mod->oid);

   if (ipc_get_clients (config->socket,&clients) < 0)
	 return (-1);

   do
	 {
		if (add_uptime (mod,clients.uptime) < 0 || add_client_number (mod,clients.n) < 0)
		  break;

		result = 0;

		for (i = 0; i < clients.n; i++)
		  if (add_client (mod,clients.info + i,clients.counter + i,i + 1) < 0)
			{
			   result = -1;
			   break;
			}
	 }
   while (0);

   ipc_destroy_clients (&clients);

   if (result < 0) odb_remove (mod->oid);

   return (result);
}

static int traffic_open (const struct mod_r *mod,const char *confdir)
{
   if ((config = config_load (mod->name,confdir)) == NULL)
	 return (-1);

   return (0);
}

static void traffic_close (const struct mod_r *mod)
{
   config_close ();
}

/* iso.org.dod.internet.private.enterprises.frogfoot.traffic */
static const uint32_t traffic_oid[8] = { 7, 43, 6, 1, 4, 1, 10002, 2 };

const mod_t mod[] =
{
   { "frogfoot/traffic", traffic_open, traffic_update, traffic_close, traffic_oid },
   { NULL, NULL, NULL, NULL, NULL }
};

