
/*
 * Copyright (c) Abraham vd Merwe <abz@blio.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	  notice, this list of conditions and the following disclaimer in the
 *	  documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of other contributors
 *	  may be used to endorse or promote products derived from this software
 *	  without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>

#include <debug/memory.h>
#include <debug/log.h>

#include <config/parse.h>

#include <tinysnmp/compat/inttypes.h>

#include <ipc.h>

#include "config.h"

#define ARRAYSIZE(x) (sizeof (x) / sizeof ((x)[0]))

#define CONFIGFILE	"traffic.conf"

/* be careful not to call these functions more than 4294967295 times (; */
static uint32_t used = 0;

/* where we keep all the goodies */
static traffic_t priv;

/* list of protocols in global scope */
static const stmt_t *protocols;

static const stmt_t *find_statement (const stmt_t *s,const char *name)
{
   const stmt_t *tmp,*prev = s;

   if (s == NULL) return (NULL);

   do
	 {
		tmp = s->next;
		if (!strcmp (s->name,name)) return (s);
		s = tmp;
	 }
   while (prev != s);

   return (NULL);
}

static int isvalid (const char *s)
{
   static const char *supported[] = { "smtp", "dns", "http", "pop3", "ftp", "ssh", "telnet", "tftp", "imap", "snmp", "irc" };
   int i;

   for (i = 0; i < ARRAYSIZE (supported); i++)
	 if (!strcmp (s,supported[i]))
	   return (1);

   return (0);
}

static protocol_t getproto (const char *s)
{
   if (!strcmp (s,"smtp")) return (smtp);
   if (!strcmp (s,"dns")) return (dns);
   if (!strcmp (s,"http")) return (http);
   if (!strcmp (s,"pop3")) return (http);
   if (!strcmp (s,"ftp")) return (ftp);
   if (!strcmp (s,"ssh")) return (ssh);
   if (!strcmp (s,"telnet")) return (telnet);
   if (!strcmp (s,"tftp")) return (tftp);
   if (!strcmp (s,"imap")) return (imap);
   if (!strcmp (s,"snmp")) return (snmp);
   if (!strcmp (s,"irc")) return (irc);
   log_printf (LOG_ERROR,"BUG in file %s on line %d: unsupported port wasn't caught by validity check!\n",__FILE__,__LINE__);
   return (unknown);
}

static int parse_protocols (const stmt_t **d,const stmt_t *s)
{
   static const char name[] = "protocols";
   int i;

   *d = find_statement (s,name);

   if (*d == NULL) return (0);

   if (!(*d)->n)
	 {
		log_printf (LOG_ERROR,"no protocols defined in ``%s'' array!\n",name);
		return (-1);
	 }

   if ((*d)->type != TOK_ENUM)
	 {
		log_printf (LOG_ERROR,"``%s'' value is not an array of enumerations!\n",name);
		return (-1);
	 }

   for (i = 0; i < (*d)->n; i++)
	 if (!isvalid ((*d)->args[i].string))
	   {
		  log_printf (LOG_ERROR,"entry ``%s'' in ``%s'' array is not a supported protocol type!\n",(*d)->args[i].string,name);
		  return (-1);
	   }

   return (0);
}

static char *parse_string (const stmt_t *stmt,const char *name)
{
   char *str;

   if ((stmt = find_statement (stmt,name)) == NULL)
	 {
		log_printf (LOG_ERROR,"no ``%s'' statement found!\n",name);
		return (NULL);
	 }

   if (stmt->type != TOK_STRING || stmt->n != 1)
	 {
		log_printf (LOG_ERROR,"``%s'' value is not a string!\n",name);
		return (NULL);
	 }

   if ((str = (char *) mem_alloc ((strlen (stmt->args->string) + 1) * sizeof (char))) == NULL)
	 {
		log_printf (LOG_ERROR,"couldn't allocate memory for temporary storage: %s\n",strerror (errno));
		return (NULL);
	 }

   strcpy (str,stmt->args->string);

   return (str);
}

static int parse_uid (const stmt_t *stmt,const char *name,uid_t *uid)
{
   struct passwd *pw;

   if ((stmt = find_statement (stmt,name)) == NULL)
	 {
		log_printf (LOG_ERROR,"no ``%s'' statement found!\n",name);
		return (-1);
	 }

   if (stmt->n != 1 || (stmt->type != TOK_STRING && stmt->type != TOK_INTEGER))
	 {
		log_printf (LOG_ERROR,"no ``%s'' string/integer found!\n",name);
		return (-1);
	 }

   if (stmt->type == TOK_INTEGER)
	 {
		if (stmt->args->integer < 0)
		  {
			 log_printf (LOG_ERROR,"``%s'' value out of range!\n",name);
			 return (-1);
		  }

		*uid = stmt->args->integer;

		return (0);
	 }

   if ((pw = getpwnam (stmt->args->string)) == NULL)
	 {
		log_printf (LOG_ERROR,"couldn't convert user name %s to a uid: %s\n",stmt->args->string,strerror (errno));
		return (-1);
	 }

   *uid = pw->pw_uid;

   return (0);
}

static int parse_gid (const stmt_t *stmt,const char *name,gid_t *gid)
{
   struct group *gr;

   if ((stmt = find_statement (stmt,name)) == NULL)
	 {
		log_printf (LOG_ERROR,"no ``%s'' statement found!\n",name);
		return (-1);
	 }

   if (stmt->n != 1 || (stmt->type != TOK_STRING && stmt->type != TOK_INTEGER))
	 {
		log_printf (LOG_ERROR,"no ``%s'' string/integer found!\n",name);
		return (-1);
	 }

   if (stmt->type == TOK_INTEGER)
	 {
		if (stmt->args->integer < 0)
		  {
			 log_printf (LOG_ERROR,"``%s'' value out of range!\n",name);
			 return (-1);
		  }

		*gid = stmt->args->integer;

		return (0);
	 }

   if ((gr = getgrnam (stmt->args->string)) == NULL)
	 {
		log_printf (LOG_ERROR,"couldn't convert group name %s to a uid: %s\n",stmt->args->string,strerror (errno));
		return (-1);
	 }

   *gid = gr->gr_gid;

   return (0);
}

static char *parse_enum (const stmt_t *stmt,const char *name)
{
   char *str;

   if ((stmt = find_statement (stmt,name)) == NULL)
	 {
		log_printf (LOG_ERROR,"no ``%s'' statement found!\n",name);
		return (NULL);
	 }

   if (stmt->type != TOK_ENUM || stmt->n != 1)
	 {
		log_printf (LOG_ERROR,"``%s'' value is not an enumeration!\n",name);
		return (NULL);
	 }

   if ((str = (char *) mem_alloc ((strlen (stmt->args->string) + 1) * sizeof (char))) == NULL)
	 {
		log_printf (LOG_ERROR,"couldn't allocate memory for temporary storage: %s\n",strerror (errno));
		return (NULL);
	 }

   strcpy (str,stmt->args->string);

   return (str);
}

static int parse_id (const stmt_t *stmt,const char *name,uint32_t *result)
{
   if ((stmt = find_statement (stmt,name)) == NULL)
	 {
		log_printf (LOG_ERROR,"no ``%s'' statement found!\n",name);
		return (-1);
	 }

   if (stmt->type != TOK_INTEGER || stmt->n != 1)
	 {
		log_printf (LOG_ERROR,"``%s'' value is not an integer!\n",name);
		return (-1);
	 }

   if (stmt->args->integer < 1 || stmt->args->integer > 2147483647)
	 {
		log_printf (LOG_ERROR,"``%s'' out of range. should be a value between 1 and 2147483647 inclusive!\n",name);
		return (-1);
	 }

   *result = stmt->args->integer;

   return (0);
}

static int parse_uint32 (const stmt_t *stmt,const char *name,uint32_t *result)
{
   if ((stmt = find_statement (stmt,name)) == NULL)
	 {
		log_printf (LOG_ERROR,"no ``%s'' statement found!\n",name);
		return (-1);
	 }

   if (stmt->type != TOK_INTEGER || stmt->n != 1)
	 {
		log_printf (LOG_ERROR,"``%s'' value is not an integer!\n",name);
		return (-1);
	 }

   if (stmt->args->integer < 0 || stmt->args->integer > 4294967)
	 {
		log_printf (LOG_ERROR,"``%s'' out of range. should be a value between 0 and 4294967 inclusive!\n",name);
		return (-1);
	 }

   *result = stmt->args->integer * 1000;

   return (0);
}

static int parse_uint16 (const stmt_t *stmt,const char *name,uint16_t *result)
{
   if ((stmt = find_statement (stmt,name)) == NULL)
	 {
		log_printf (LOG_ERROR,"no ``%s'' statement found!\n",name);
		return (-1);
	 }

   if (stmt->type != TOK_INTEGER || stmt->n != 1)
	 {
		log_printf (LOG_ERROR,"``%s'' value is not an integer!\n",name);
		return (-1);
	 }

   if (stmt->args->integer < 0 || stmt->args->integer > 0xffff)
	 {
		log_printf (LOG_ERROR,"``%s'' out of range. should be a value between 0 and 65535 inclusive!\n",name);
		return (-1);
	 }

   *result = stmt->args->integer;

   return (0);
}

static int parse_client (const char *config,client_t *client,const section_t *section)
{
   char *tmp;
   const stmt_t *stmt;
   uint32_t i;

   if (parse_id (section->stmt,"id",&client->info.id) < 0)
	 {
		err0:
		log_printf (LOG_ERROR,"parse error in client section %s of %s!\n",section->name,config);
		return (-1);
	 }

   if ((client->info.name = parse_string (section->stmt,"name")) == NULL)
	 goto err0;

   if ((tmp = parse_enum (section->stmt,"quality")) == NULL)
	 {
		err1:
		mem_free (client->info.name);
		goto err0;
	 }

   if (!strcmp (tmp,"bronze"))
	 client->info.quality = IPC_QUALITY_BRONZE;
   else if (!strcmp (tmp,"silver"))
	 client->info.quality = IPC_QUALITY_SILVER;
   else if (!strcmp (tmp,"gold"))
	 client->info.quality = IPC_QUALITY_GOLD;
   else if (!strcmp (tmp,"platinum"))
	 client->info.quality = IPC_QUALITY_PLATINUM;
   else
	 {
		log_printf (LOG_ERROR,"quality should be bronze, silver, gold, or platinum\n");
		mem_free (tmp);
		goto err1;
	 }

   mem_free (tmp);

   if (parse_uint32 (section->stmt,"input",&client->info.input) < 0 ||
	   parse_uint32 (section->stmt,"output",&client->info.output) < 0)
	 goto err1;

   if ((stmt = find_statement (section->stmt,"networks")) == NULL)
	 {
		log_printf (LOG_ERROR,"no ``networks'' statement found!\n");
		goto err1;
	 }

   if (stmt->type != TOK_STRING || stmt->n < 1)
	 {
		log_printf (LOG_ERROR,"``networks'' is not a string or no values defined!\n");
		goto err1;
	 }

   client->info.n = stmt->n;

   if ((client->info.networks = (network_t *) mem_alloc (sizeof (network_t) * client->info.n)) == NULL)
	 {
		log_printf (LOG_ERROR,"couldn't allocate memory for temporary storage: %s\n",strerror (errno));
		goto err1;
	 }

   for (i = 0; i < client->info.n; i++)
	 if (aton (client->info.networks + i,stmt->args[i].string) < 0)
	   {
		  log_printf (LOG_ERROR,"invalid network ``%s'' defined in list of allowed networks!\n",stmt->args[i].string);
		  mem_free (client->info.networks);
		  goto err1;
	   }

   if (parse_protocols (&stmt,section->stmt) < 0 || (stmt == NULL && protocols == NULL))
	 goto err1;

   if (stmt == NULL) stmt = protocols;

   client->n = stmt->n + 2;

   if ((client->recv = (client_counter_t *) mem_alloc (client->n * sizeof (client_counter_t))) == NULL)
	 {
		log_printf (LOG_ERROR,"couldn't allocate memory for temporary storage: %s\n",strerror (errno));
		goto err1;
	 }

   if ((client->xmit = (client_counter_t *) mem_alloc (client->n * sizeof (client_counter_t))) == NULL)
	 {
		log_printf (LOG_ERROR,"couldn't allocate memory for temporary storage: %s\n",strerror (errno));
		mem_free (client->recv);
		goto err1;
	 }

   /* the first two entries should always be a one-to-one mapping */
   client->recv[unknown].proto = client->xmit[unknown].proto = unknown;
   client->recv[total].proto = client->xmit[total].proto = total;

   for (i = 2; i < client->n; i++)
	 client->recv[i].proto = client->xmit[i].proto = getproto (stmt->args[i - 2].string);

   return (0);
}

static int parse_config (const char *config,const section_t *section)
{
   uint32_t i,j;

   if (!section->n)
	 {
		log_printf (LOG_ERROR,"no client sections found in %s!\n",config);
		return (-1);
	 }

   if ((priv.interface = parse_string (section->stmt,"interface")) == NULL)
	 return (-1);

   if ((priv.pidfile = parse_string (section->stmt,"pidfile")) == NULL)
	 {
		mem_free (priv.interface);
		return (-1);
	 }

   if ((priv.socket = parse_string (section->stmt,"socket")) == NULL)
	 {
		mem_free (priv.interface);
		mem_free (priv.pidfile);
		return (-1);
	 }

   if (parse_uid (section->stmt,"user",&priv.uid) < 0 || parse_gid (section->stmt,"group",&priv.gid) < 0)
	 {
		mem_free (priv.interface);
		mem_free (priv.pidfile);
		mem_free (priv.socket);
		return (-1);
	 }

   if (parse_uint16 (section->stmt,"rulestart",&priv.rulestart) < 0 || parse_protocols (&protocols,section->stmt) < 0)
	 {
		mem_free (priv.interface);
		mem_free (priv.pidfile);
		mem_free (priv.socket);
		return (-1);
	 }

   priv.n = section->n;

   if ((priv.clients = (client_t *) mem_alloc (sizeof (client_t) * priv.n)) == NULL)
	 {
		log_printf (LOG_ERROR,"couldn't allocate memory for temporary storage: %s\n",strerror (errno));
		mem_free (priv.interface);
		mem_free (priv.pidfile);
		mem_free (priv.socket);
		return (-1);
	 }

   for (i = 0; i < priv.n; i++)
	 {
		if (parse_client (config,priv.clients + i,section->child[i]) < 0)
		  {
			 client_failure:
			 priv.n = i;
			 for (i = 0; i < priv.n; i++)
			   {
				  mem_free (priv.clients[i].info.name);
				  mem_free (priv.clients[i].info.networks);
				  mem_free (priv.clients[i].recv);
				  mem_free (priv.clients[i].xmit);
			   }
			 mem_free (priv.clients);
			 mem_free (priv.interface);
			 mem_free (priv.pidfile);
			 mem_free (priv.socket);
			 return (-1);
		  }

		for (j = 0; j < i; j++)
		  if (priv.clients[i].info.id == priv.clients[j].info.id)
			{
			   log_printf (LOG_ERROR,"duplicate id %u for client %s\n",priv.clients[i].info.id,priv.clients[i].info.name);
			   i++;
			   goto client_failure;
			}
	 }

   return (0);
}

/*
 * free memory allocated by the config_load function.
 */
static void config_close ()
{
   if (used)
	 {
		/* last guy out close the door */
		if (used == 1)
		  {
			 uint32_t i;
			 mem_free (priv.interface);
			 mem_free (priv.pidfile);
			 mem_free (priv.socket);
			 for (i = 0; i < priv.n; i++)
			   {
				  mem_free (priv.clients[i].info.name);
				  mem_free (priv.clients[i].info.networks);
				  mem_free (priv.clients[i].recv);
				  mem_free (priv.clients[i].xmit);
			   }
			 mem_free (priv.clients);
		  }

		/* decrement usage count */
		used--;
	 }
}

/*
 * parse configuration file. return 0 if successful, -1 if some error
 * occurred. the function automatically logs all errors.
 */
int config_load (traffic_t *traffic,const char *path)
{
   /* increment usage count */
   used++;

   /* if its the first call, we have to load the configuration file */
   if (used == 1)
	 {
		section_t *section;

		if ((section = parse (path)) == NULL)
		  return (-1);

		if (parse_config (path,section) < 0)
		  {
			 parse_destroy (&section);
			 return (-1);
		  }

		*traffic = priv;

		parse_destroy (&section);
	 }

   atexit (config_close);

   return (0);
}

