
/*
 * Copyright (c) Abraham vd Merwe <abz@blio.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	  notice, this list of conditions and the following disclaimer in the
 *	  documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of other contributors
 *	  may be used to endorse or promote products derived from this software
 *	  without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/ioctl.h>
#include <sys/queue.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netinet/ip_fw.h>
#include <net/if.h>

#include <debug/log.h>
#include <debug/memory.h>

#include <tinysnmp/compat/inttypes.h>

#include "aton.h"
#include "fw.h"

static int raw_socket_fd = -1;

static void close_raw_socket ()
{
   close (raw_socket_fd);
}

static int open_raw_socket (int level)
{
   if (raw_socket_fd > 0)
	 return (raw_socket_fd);

   if ((raw_socket_fd = socket (PF_INET,SOCK_RAW,IPPROTO_RAW)) < 0)
	 {
		log_printf (level,"socket: %s\n",strerror (errno));
		return (-1);
	 }

   atexit (close_raw_socket);

   return (raw_socket_fd);
}

/*
 * Add rules to count traffic that matches the specified criteria. Returns 0
 * if successful, -1 otherwise. The function automatically logs errors.
 *
 *        level     Log level
 *        n         Rule number
 *        proto     IP protocol (IPPROTO_IP, IPPROTO_TCP, or IPPROTO_UDP)
 *        net       Source/Destination network to match
 *        port      Port number (if 0, any port matches)
 *        iface     Name of interface on which to match packets
 *        src       If 0, match destination, else match source
 */
int fw_add_counter (int level,uint16_t n,int proto,const network_t *net,uint16_t port,const char *iface,int src)
{
   struct ip_fw rule;
   char *s,*endptr;
   unsigned long unit;
   int fd;
   struct ifreq ifr;
   socklen_t len;

   memset (&rule,0,sizeof (rule));

   rule.fw_number = n;
   rule.fw_flg = IP_FW_F_COUNT | IP_FW_F_IIFNAME | IP_FW_F_OIFNAME | IP_FW_F_OUT | IP_FW_F_IN | IP_FW_F_IIFACE | IP_FW_F_OIFACE;
   rule.fw_prot = proto;

   strncpy (rule.fw_out_if.fu_via_if.name,iface,sizeof (rule.fw_out_if.fu_via_if.name));
   rule.fw_out_if.fu_via_if.name[sizeof (rule.fw_out_if.fu_via_if.name) - 1] = '\0';
   s = rule.fw_out_if.fu_via_if.name;
   while (*s != '\0' && !isdigit (*s)) s++;

   unit = strtoul (s,&endptr,10);

   if (*s == '\0' || *endptr != '\0' || (unit == ULONG_MAX && errno == ERANGE) || unit > SHRT_MAX)
	 {
		log_printf (level,"``%s'' isn't a valid interface\n",rule.fw_out_if.fu_via_if.name);
		return (-1);
	 }

   *s = '\0';

   rule.fw_out_if.fu_via_if.unit = unit;
   rule.fw_in_if = rule.fw_out_if;

   if ((fd = open_raw_socket (level)) < 0)
	 return (-1);

   snprintf (ifr.ifr_name,
			 sizeof (ifr.ifr_name),
			 "%s%u",
			 rule.fw_out_if.fu_via_if.name,
			 rule.fw_out_if.fu_via_if.unit);
   ifr.ifr_name[sizeof (ifr.ifr_name) - 1] = '\0';

   if (ioctl (fd,SIOCGIFFLAGS,&ifr) < 0)
	 {
		log_printf (level,"ioctl SIOCGIFFLAGS: %s (does interface ``%s'' exist?)\n",strerror (errno),ifr.ifr_name);
		return (-1);
	 }

   if (src)
	 {
		rule.fw_src.s_addr = net->address;
		rule.fw_smsk.s_addr = net->netmask;
	 }
   else
	 {
		rule.fw_dst.s_addr = net->address;
		rule.fw_dmsk.s_addr = net->netmask;
	 }

   if (port)
	 {
		rule.fw_uar.fw_pts[0] = port;
		IP_FW_SETNSRCP (&rule,1);
	 }

again:
   len = sizeof (struct ip_fw);

   if (getsockopt (fd,IPPROTO_IP,IP_FW_ADD,&rule,&len) < 0)
	 {
		log_printf (level,"getsockopt IP_FW_ADD: %s\n",strerror (errno));
		return (-1);
	 }

   if (port)
	 {
		IP_FW_SETNSRCP (&rule,0);
		IP_FW_SETNDSTP (&rule,1);
		port = 0;
		goto again;
	 }

   return (0);
}

/*
 * Delete rules that match the specified rule number. Returns 0 if
 * successful, -1 otherwise. The function automatically logs errors.
 *
 *        level     Log level
 *        n         Rule number
 */
int fw_del_counters (int level,uint16_t n)
{
   struct ip_fw rule;
   int fd;

   rule.fw_number = n;

   if ((fd = open_raw_socket (level)) < 0)
	 return (-1);

   if (setsockopt (fd,IPPROTO_IP,IP_FW_DEL,&rule,sizeof (struct ip_fw)) < 0)
	 {
		log_printf (level,"setsockopt IP_FW_DEL: %s\n",strerror (errno));
		return (-1);
	 }

   return (0);
}

/*
 * Get a snapshot of the current firewall rules. This is needed by
 * fw_get_counters(). Returns a pointer to the allocated data structure
 * of NULL if some error occurred. The function automatically logs errors.
 *
 *        level     Log level
 */
void *fw_get_rules (int level)
{
   int fd;
   struct ip_fw *ptr,*rules = NULL;
   socklen_t olen,ilen = 0;

   if ((fd = open_raw_socket (level)) < 0)
	 return (NULL);

   do
	 {
		ilen += sizeof (struct ip_fw);

		if ((ptr = (struct ip_fw *) mem_realloc ((void *) rules,ilen)) == NULL)
		  {
			 log_printf (level,"realloc: %s\n",strerror (errno));
			 if (rules != NULL) mem_free (rules);
			 return (NULL);
		  }

		rules = ptr;
		olen = ilen;

		if (getsockopt (fd,IPPROTO_IP,IP_FW_GET,rules,&olen) < 0)
		  {
			 log_printf (level,"getsockopt IP_FW_GET: %s\n",strerror (errno));
			 mem_free (rules);
			 return (NULL);
		  }
	 }
   while (olen == ilen);

   return ((void *) rules);
}

/*
 * Return the sum of the interface byte count counters of the rules
 * that match the specified rule number. Returns 0 if one of more
 * rules matched, or -1 if no firewall rules match the specified
 * rule number.
 *
 *        rules     List of rules allocated by fw_get_rules()
 *        n         Rule number
 *        count     Sum of matched rule byte counters
 */
int fw_get_counters (void *rules,uint16_t n,uint64_t *count)
{
   uint16_t i = 0;
   struct ip_fw *ipfw_rules = (struct ip_fw *) rules;
   int result = -1;

   *count = 0;

   do
	 {
		result = 0;
		if (ipfw_rules[i].fw_number == n)
		  *count += ipfw_rules[i].fw_bcnt;
	 }
   while (ipfw_rules[i++].fw_number < 65535);

   return (result);
}

/*
 * Free memory allocated by the fw_get_rules() function.
 *
 *        rules     List of rules allocated by fw_get_rules()
 */
void fw_destroy_rules (void *rules)
{
   if (rules != NULL) mem_free (rules);
}

