
/*
 * Copyright (c) Abraham vd Merwe <abz@blio.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	  notice, this list of conditions and the following disclaimer in the
 *	  documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of other contributors
 *	  may be used to endorse or promote products derived from this software
 *	  without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/un.h>
#include <sys/socket.h>
#include <unistd.h>
#include <fcntl.h>

#include <debug/log.h>
#include <debug/memory.h>

#include <tinysnmp/compat/inttypes.h>

#include "config.h"
#include "rules.h"
#include "ipc.h"

#define MAX(a,b) ({						\
		const typeof (a) _a = (a);		\
		const typeof (b) _b = (b);		\
		(void) (&_a == &_b);			\
		_a > _b ? _a : _b;				\
	})

typedef struct
{
   const char *logfile;					/* location of log file				*/
   int loglevel;						/* logging system verbosity			*/
   int daemon;							/* are we a daemon?					*/
   const char *configfile;				/* location of configuration file	*/
} options_t;

#define usage(p) help_stub(p,0)
#define help(p) help_stub(p,1)
static void help_stub (const char *progname,int verbose)
{
   fprintf (stderr,
			"usage: %s [OPTIONS] <configfile>\n"
			"       %s -h | --help\n",
			progname,progname);

   if (verbose)
	 {
		fprintf (stderr,
				 "\n"
				 "where OPTIONS can be any of the following:\n"
				 "\n"
				 "%s"
				 "\n"
				 "    -f | --logfile=FILENAME     Location of log file. By default, messages will be\n"
				 "                                logged to the console.\n"
				 "\n"
				 "    -l | --loglevel=LEVEL       The following levels can be specified (in increasing\n"
				 "                                order of verbosity):\n"
				 "\n"
				 "                                    quiet     No messages will be logged\n"
				 "                                    errors    Only errors will be logged\n"
				 "                                    warnings  Warnings & errors will be logged\n"
				 "                                    normal    Default messages and all of the above\n"
				 "                                    verbose   Verbose messages will also be logged\n"
				 "                                    debug     Log debug messages as well\n"
				 "                                    noisy     Noisy debug messages will be logged as well\n"
				 "\n"
				 "                                If a log file is specified, ORCA defaults to verbose'',\n"
				 "                                otherwise, noisy'' is selected.\n"
				 "\n"
				 "    -d | --daemon               Run in the background. You need to specify a log file if\n"
				 "                                you want to use this option.\n"
				 "\n"
				 "    -h | --help                 Show this help message.\n"
				 "\n"
				 "    <configfile>                Location of ORCA configuration file.\n"
				 "\n",
				 progname);
	 }

   exit (EXIT_FAILURE);
}

static void parse_options (options_t *opts,int argc,char *argv[])
{
   const char *progname;
   int i;

   (progname = strrchr (argv[0],'/')) ? progname++ : (progname = argv[0]);

   /* is the user requesting help? */
   if (argc == 2 && (!strcmp (argv[1],"-h") || !strcmp (argv[1],"--help")))
	 help (progname);

   if (argc < 2) usage (progname);

   opts->logfile = NULL;
   opts->loglevel = -1;
   opts->daemon = 0;

   for (i = 1; i < argc - 1; i++)
	 {
		if (!strcmp (argv[i],"-d") || !strcmp (argv[i],"--daemon"))
		  opts->daemon = 1;
		else if (!strcmp (argv[i],"-f") || !strncmp (argv[i],"--logfile=",10))
		  {
			 char *arg = strlen (argv[i]) == 2 ? argv[++i] : strchr (argv[i],'=') + 1;
			 if (i == argc - 1 || *arg == '\0') usage (progname);
			 opts->logfile = arg;
		  }
		else if (!strcmp (argv[i],"-l") || !strncmp (argv[i],"--loglevel=",11))
		  {
			 char *arg = strlen (argv[i]) == 2 ? argv[++i] : strchr (argv[i],'=') + 1;
			 if (i == argc - 1 || *arg == '\0') usage (progname);

			 if (!strcmp (arg,"quiet"))
			   opts->loglevel = LOG_QUIET;
			 else if (!strcmp (arg,"errors"))
			   opts->loglevel = LOG_ERROR;
			 else if (!strcmp (arg,"warnings"))
			   opts->loglevel = LOG_WARNING;
			 else if (!strcmp (arg,"normal"))
			   opts->loglevel = LOG_NORMAL;
			 else if (!strcmp (arg,"verbose"))
			   opts->loglevel = LOG_VERBOSE;
			 else if (!strcmp (arg,"debug"))
			   opts->loglevel = LOG_DEBUG;
			 else if (!strcmp (arg,"noisy"))
			   opts->loglevel = LOG_NOISY;
			 else usage (progname);
		  }
		else usage (progname);
	 }

   opts->configfile = argv[i];
   if (opts->daemon && opts->logfile == NULL)
	 usage (progname);

   if (opts->loglevel == -1)
	 opts->loglevel = opts->logfile == NULL ? LOG_NOISY : LOG_VERBOSE;
}

static int listen_fd = -1,client_fd = -1;
static traffic_t traffic;
static volatile int finished = 0;
static volatile int rotated = 0;

static void catch_signals (int sig)
{
   if (sig == SIGUSR1)
	 rotated = 1;
   else
	 finished = 1;
}

static void cleanup1 ()
{
   mem_close ();
   log_close ();
}

static void cleanup2 ()
{
   /* delete firewall rules */
   rules_remove (&traffic);

   /* close socket(s) */
   if (listen_fd > 0) close (listen_fd);
   if (client_fd > 0) close (client_fd);

   /* remove socket */
   unlink (traffic.socket);
}

static int create_pidfile (const char *filename)
{
   int fd,result;
   char str[32];

   if ((fd = creat (filename,0644)) < 0) return (-1);

   sprintf (str,"%u\n",getpid ());

   if ((result = write (fd,str,strlen (str))) != strlen (str))
	 {
		int saved = result < 0 ? errno : EIO;
		close (fd);
		errno = saved;
		return (-1);
	 }

   close (fd);

   return (0);
}

static void fail (const char *fmt, ...)
{
   va_list ap;
   va_start (ap,fmt);
   if (fmt != NULL) log_vprintf (LOG_ERROR,fmt,ap);
   va_end (ap);
   exit (EXIT_FAILURE);
}

int main (int argc,char *argv[])
{
   options_t options;
   fd_set rfds;
   struct timeval tv;
   int accept_fd,result;
   uint32_t cmd,client;
   struct sockaddr_un addr;

   /* ignore the important signals */
   signal (SIGHUP,SIG_IGN);
   signal (SIGINT,SIG_IGN);
   signal (SIGTERM,SIG_IGN);
   signal (SIGUSR1,SIG_IGN);
   signal (SIGUSR2,SIG_IGN);
   signal (SIGTSTP,SIG_IGN);

   /* parse command-line options */
   parse_options (&options,argc,argv);

   /* initialize memory detection and logging subsystems */
   mem_open (fail);
   if (log_open (options.logfile,options.loglevel,options.logfile == NULL ? LOG_HAVE_COLORS : LOG_DEBUG_PREFIX_ONLY) < 0)
	 {
		mem_close ();
		fail ("unable to initialize logging system: %s\n",strerror (errno));
	 }
   atexit (cleanup1);

   /* parse configuration file */
   if (config_load (&traffic,options.configfile) < 0)
	 fail (NULL);

   /* add firewall rules */
   if (rules_add (&traffic) < 0)
	 fail (NULL);
   atexit (cleanup2);

   /* remove stale socket(s), if any */
   unlink (traffic.socket);

   /* create a unix socket */
   if ((listen_fd = socket (PF_UNIX,SOCK_STREAM,0)) < 0)
	 fail ("unable to create socket: %s\n",strerror (errno));

   addr.sun_family = AF_UNIX;
   strncpy (addr.sun_path,traffic.socket,sizeof (addr.sun_path));
   addr.sun_path[sizeof (addr.sun_path) - 1] = '\0';

   /* bind the socket to a name */
   if (bind (listen_fd,(struct sockaddr *) &addr,SUN_LEN (&addr)) < 0)
	 fail ("unable to bind socket %s: %s\n",addr.sun_path,strerror (errno));

   /* change fifo ownership */
   if (chown (traffic.socket,traffic.uid,traffic.gid) < 0)
	 fail ("unable to change ownership of socket %s to %u:%u: %s\n",addr.sun_path,traffic.uid,traffic.gid,strerror (errno));

   /* listen for connections. we don't care about pending connections. */
   if (listen (listen_fd,0) < 0)
	 fail ("unable to listen for connections on socket %s: %s\n",addr.sun_path,strerror (errno));

   if (options.daemon)
	 {
		/* fork and become a daemon */
		if (daemon (0,0) < 0)
		  fail ("unable to daemonize myself: %s\n",strerror (errno));

		/* now create the pidfile */
		if (create_pidfile (traffic.pidfile) < 0)
		  fail ("unable to create pid file %s: %s\n",traffic.pidfile,strerror (errno));
	 }

   /* catch all the signals we're interested in */
   signal (SIGINT,catch_signals);
   signal (SIGTERM,catch_signals);
   signal (SIGUSR1,catch_signals);

   /* once we enter the main loop, we try our best not to abort */
   while (!finished)
	 {
		/* reset the log file */
		if (rotated)
		  {
			 if (log_reset () < 0) break;
			 rotated = 0;
		  }

		/* listen for incoming connections */
		FD_ZERO (&rfds);
		FD_SET (listen_fd,&rfds);
		if (client_fd > 0) FD_SET (client_fd,&rfds);

		/* wait 5 minutes */
		tv.tv_sec = 300;
		tv.tv_usec = 0;

		result = select (MAX (listen_fd,client_fd) + 1,&rfds,NULL,NULL,&tv);

		/* did something bad happen? */
		if (result < 0 && errno != EINTR && errno != ENOMEM)
		  fail ("select failed: %s\n",strerror (errno));

		/* time to do some work */
		if (result > 0)
		  {
			 /* accept pending connections */
			 if (FD_ISSET (listen_fd,&rfds))
			   {
				  socklen_t len = SUN_LEN (&addr);
				  if ((accept_fd = accept (listen_fd,(struct sockaddr *) &addr,&len)) < 0)
					log_printf (LOG_WARNING,"unable to accept pending connection\n");
				  else if (client_fd < 0)
					client_fd = accept_fd;
				  else close (accept_fd);
			   }

			 /* talk to client */
			 if (client_fd > 0 && FD_ISSET (client_fd,&rfds))
			   {
				  result = ipc_get (client_fd,&cmd,&client);

				  if (!result)
					switch (cmd)
					  {
					   case IPC_PACKET_UPDATE:
						 log_printf (LOG_DEBUG,"update\n");
						 rules_update (&traffic);
						 break;

					   case IPC_PACKET_CLIENTS:
						 log_printf (LOG_DEBUG,"clients\n");
						 result = ipc_put_clients (client_fd,traffic.n);
						 break;

					   case IPC_PACKET_INFO:
						 log_printf (LOG_DEBUG,"info %" PRIu32 "\n",client);
						 if (client >= traffic.n)
						   {
							  result = -1;
							  log_printf (LOG_WARNING,"illegal client index %" PRIu32 "\n",client);
						   }
						 else result = ipc_put_info (client_fd,traffic.clients + client);
						 break;

					   case IPC_PACKET_COUNTERS:
						 log_printf (LOG_DEBUG,"counters %" PRIu32 "\n",client);
						 if (client >= traffic.n)
						   {
							  result = -1;
							  log_printf (LOG_WARNING,"illegal client index %" PRIu32 "\n",client);
						   }
						 else result = ipc_put_counters (client_fd,traffic.clients + client);
						 break;

					   case IPC_PACKET_UPTIME:
						 log_printf (LOG_DEBUG,"uptime\n");
						 result = ipc_put_uptime (client_fd,traffic.uptime);
						 break;
					  }

				  /* client closed the connection */
				  close (client_fd);
				  client_fd = -1;
			   }
		  }
	 }

   exit (EXIT_SUCCESS);
}

