
/*
 * Copyright (c) Abraham vd Merwe <abz@blio.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	  notice, this list of conditions and the following disclaimer in the
 *	  documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of other contributors
 *	  may be used to endorse or promote products derived from this software
 *	  without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <net/if.h>

#include <debug/memory.h>
#include <debug/log.h>

#include <tinysnmp/compat/inttypes.h>

#include <abz/getline.h>
#include <abz/trim.h>
#include <abz/atou32.h>

#include "proc.h"

#define _PATH_PROCNET_DEV "/proc/net/dev"

#define NB_OF(x) (sizeof (x) / sizeof ((x)[0]))

static int getifstats (devstat_t *dev,char *proc)
{
   uint32_t *var[] =
	 {
		&dev->rx_bytes,
		&dev->rx_packets,
		&dev->rx_errors,
		&dev->rx_dropped,
		&dev->rx_fifo_errors,
		&dev->rx_frame_errors,
		&dev->rx_compressed,
		&dev->multicast,
		&dev->tx_bytes,
		&dev->tx_packets,
		&dev->tx_errors,
		&dev->tx_dropped,
		&dev->tx_fifo_errors,
		&dev->collisions,
		&dev->tx_carrier_errors,
		&dev->tx_compressed
	 };
   char *d,*s = proc;
   int i;

   for (i = 0; i < NB_OF (var); i++)
	 {
		ltrim (s);
		if ((d = strchr (s,' ')) != NULL) *d++ = '\0';
		if (atou32 (s,var[i]) < 0) return (-1);
		s = d;
	 }

   if (s != NULL) return (-1);

   return (0);
}

#define getdevstats(n) getdevstats_stub(__FILE__,__LINE__,__FUNCTION__,n)
devstat_t *getdevstats_stub (const char *filename,int line,const char *function,size_t *n)
{
   int i,fd,saved;
   char *s,*str;
   devstat_t *stats = NULL,*ptr;
   static const char *header[] =
	 {
		"Inter-|   Receive                                                |  Transmit",
		" face |bytes    packets errs drop fifo frame compressed multicast|bytes    packets errs drop fifo colls carrier compressed"
	 };

   if ((fd = open (_PATH_PROCNET_DEV,O_RDONLY)) < 0)
	 return (NULL);

   *n = 0;

   for (i = 0; (str = getline (fd)) != NULL; i++)
	 {
		if (i < 2)
		  {
			 if (strcmp (str,header[i]))
			   {
				  close (fd);
				  mem_free (str);
				  errno = EINVAL;
				  return (NULL);
			   }
		  }
		else
		  {
			 ltrim (str);
			 if ((s = strchr (str,':')) == NULL)
			   {
				  close (fd);
				  mem_free (str);
				  errno = EINVAL;
				  return (NULL);
			   }
			 *s++ = '\0';

			 if ((ptr = (devstat_t *) mem_realloc_stub (stats,(*n + 1) * sizeof (devstat_t),filename,line,function)) == NULL)
			   {
				  saved = errno;
				  error:
				  if (stats != NULL) mem_free (stats);
				  close (fd);
				  mem_free (str);
				  errno = saved;
				  return (NULL);
			   }
			 stats = ptr;

			 if (strlen (str) >= IFNAMSIZ || getifstats (stats + *n,s) < 0)
			   {
				  saved = EINVAL;
				  goto error;
			   }
			 strcpy (stats[*n].dev,str);

			 (*n)++;
		  }
		mem_free (str);
	 }

   if (errno)
	 {
		saved = errno;
		if (stats != NULL) mem_free (stats);
		close (fd);
		errno = saved;
		return (NULL);
	 }

   close (fd);
   return (stats);
}

/*
 * Define this to 1 if you want to test the /proc/net/dev parser
 */

#if 0
static void fail (const char *fmt, ...)
{
   va_list ap;
   va_start (ap,fmt);
   if (fmt != NULL) log_vprintf (LOG_ERROR,fmt,ap);
   va_end (ap);
   exit (EXIT_FAILURE);
}

static void close_all ()
{
   mem_close ();
   log_close ();
}

int main ()
{
   devstat_t *stat;
   int i;
   size_t n;

   /* initialize memory detection and logging subsystems */
   mem_open (fail);
   if (log_open (NULL,LOG_NOISY,LOG_HAVE_COLORS) < 0)
	 {
		mem_close ();
		fprintf (stderr,"unable to initialize logging system: %s\n",strerror (errno));
		exit (EXIT_FAILURE);
	 }
   atexit (close_all);

   if ((stat = getdevstats (&n)) == NULL)
	 {
		log_printf (LOG_ERROR,"couldn't retrieve interface statistics from %s: %s\n",_PATH_PROCNET_DEV,strerror (errno));
		exit (EXIT_FAILURE);
	 }

   for (i = 0; i < n; i++)
	 log_printf (LOG_NORMAL,
				 "interface %s\n"
				 "   receive                   transmit\n"
				 "      bytes %-20" PRIu32 "bytes %" PRIu32 "\n"
				 "      packets %-18" PRIu32 "packets %" PRIu32 "\n"
				 "      errors %-19" PRIu32 "errors %" PRIu32 "\n"
				 "      dropped %-18" PRIu32 "dropped %" PRIu32 "\n"
				 "      fifo errors %-14" PRIu32 "fifo errors %" PRIu32 "\n"
				 "      frame errors %-13" PRIu32 "carrier errors %" PRIu32 "\n"
				 "      compressed %-15" PRIu32 "compressed %" PRIu32 "\n"
				 "      multicast %-16" PRIu32 "collisions %" PRIu32 "\n",
				 stat[i].dev,
				 stat[i].rx_bytes,stat[i].tx_bytes,
				 stat[i].rx_packets,stat[i].tx_packets,
				 stat[i].rx_errors,stat[i].tx_errors,
				 stat[i].rx_dropped,stat[i].tx_dropped,
				 stat[i].rx_fifo_errors,stat[i].tx_fifo_errors,
				 stat[i].rx_frame_errors,stat[i].tx_carrier_errors,
				 stat[i].rx_compressed,stat[i].tx_compressed,
				 stat[i].multicast,stat[i].collisions);

   mem_free (stat);

   exit (EXIT_SUCCESS);
}
#endif

